#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "tc.h"

static int EdErrCheck(estiMethod*);
static int ErrDLY(estiMethod*);
static int RefEdMean(refMethod*), RefEdExtra(refMethod*);
static real CompEdResExtra(refMethod*);
static int CompEdRes(int,int,int,int,real*);
static int AdaptEdRef(refMethod*);
static int AddErr(TD*);
static int ErrAxMulRhsDiag(int,int,int,int);
static int UnMark(TD*), RefIfEdLessTheta(TD*), RefMarked(TD*);
static void EdgeSetZero(int);
static int ErrAss(TD*,real**,real*,real*);

static int BRHS, BDIAG, BSOL , BRQ, ed_res;
static real theta;
static real **globA=nil, *globB=nil, *globMad=nil;
static integData *errIntegData=nil;
static real sigma1, sigma2;
static int ed_mad = -1;
static real sigmaxy[NO_OF_AREA_IPS], xixy[NO_OF_AREA_IPS];

static int iMark = -1;

int DefEllErrMethod()
{
   if (!DefEstiMethod("dly",EdErrCheck,ErrDLY,
                          (int(*)(estiMethod*))nil)) return false;

   if (!DefRefMethod("ed_mean",1.00,2.0,RefEdMean,AdaptEdRef,
                            (int(*)(refMethod*))nil)) return false;
   if (!DefRefMethod("ed_extra",0.5,1.0,RefEdExtra,AdaptEdRef,
                            (int(*)(refMethod*))nil)) return false;

   return true;
}

static int EdErrCheck(estiMethod *estor)
{
   if ( !(    ((nodesState->noOfPointNodes)==1)
           && ((nodesState->noOfEdgeNodes)==0)
           && ((nodesState->noOfTriangleNodes)==0)
           && ((nodesState->noOfTetrahedronNodes)==0))  )
     {
       ZIBStdOut("ellkask error estimator not suited for this type of problem\n");
       return false;
     }
   if ((estor->ed_res)==-1)
     estor->ed_res =
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"res");
   if ((estor->ed_rq)==-1)
     estor->ed_rq = 
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rq");
   if ((estor->ed_diag)==-1)
     estor->ed_diag = 
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"diag");
   if ((estor->ed_rhs)==-1)
     estor->ed_rhs =
         GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"rhs");
    if (ed_mad==-1)
      ed_mad = GetAccess(nodesState->accEdge,sizeof(real),sizeof(real),"mad");

   if (((estor->ed_res)==-1)||((estor->ed_rq)==-1)||
	    ((estor->ed_diag)==-1)||((estor->ed_rhs)==-1)||
		(ed_mad==-1)) return false;

   return true;
}

static int CompEdRes(int ires, int irq, int idqq, int irhs, real *tau)
{
   EDG *ed = actTriang->firstEdge;
   double sum = ZERO, res, um;

   while (ed!=nil)
     {
       if ((ed->firstSon)!=nil) { ed = ed->next; continue; }
       RE(ed, irq) = RE(ed, irhs)-RE(ed, irq);
       if ((ed->boundP)==DIRICHLET) res = ZERO;
       else
         {
           um = RE(ed,irq);
            res = um/RE(ed,idqq);
			res *= res;
			res *= (1.0-(timeControl->energLmbda))*RE(ed,ed_mad)+
			             (timeControl->energLmbda)*RE(ed,idqq);
         }
       RE(ed,ires) = res;
       sum += res; 

       ed = ed->next;
     }

   if (sum<ZERO) sum = ZERO;
   *tau = sum;

   return true;
}

static int ErrDLY(estiMethod *estor)
{
   int rSol = nodesState->rSol;

   if (!ErrAxMulRhsDiag(rSol, estor->ed_rq, estor->ed_diag, estor->ed_rhs))
      return false;

   if(!(CompEdRes(estor->ed_res, estor->ed_rq,
                estor->ed_diag, estor->ed_rhs, &(estor->globError) )))
     {
       sprintf(globBuf," Error estimator failed in ErrDLY \n");
       ZIBStdOut(globBuf);
       return false;
     }
   else
	  {
	    (estor->globError) /= (timeControl->gauge);
		(estor->globError) += ((actItor->iteCount)==0)?ZERO:actItor->res;
	  }

	if ((actProblem->Sol) != nil)
	  {
		TrueError();
		if (timeControl->verboseP>1)
		  printf("ErrDLY(TrueError): %e, norm=%e, dim=%d\n",timeControl->trueError,
		         timeControl->norm,actTriang->noOfPoints);
	  }
	else
	  {
		if (timeControl->verboseP>1)
	      printf("ErrDLY(TrueError): -, norm=%e, dim=%d\n",
		         timeControl->norm,actTriang->noOfPoints);
	  }
   return true;
}

static int RefEdExtra(refMethod *refor)
{
   refor->mBar = CompEdResExtra(refor);

   return true;
}

static int RefEdMean(refMethod *refor)
{ 
   refor->mBar = (actEstor->globError)/(actTriang->noOfEdges);

   return true;
}

static real CompEdResExtra(refMethod *refor)
{
   EDG  *ed = actTriang->firstEdge;
   real sum = ZERO, res, fatherRes, cutEta = ZERO, newRes;
   int ires = actEstor->ed_res;

   while (ed!=nil)
     {
       if ((ed->firstSon)!=nil) { ed = ed->next; continue; }
       res = RE(ed,ires);

       if ((ed->father) != nil)
          {
            fatherRes = RE((ed->father),ires);
            if (fatherRes == ZERO) newRes = ZERO;
            else newRes = res*res/fatherRes;
          }
       else newRes = ZERO;

       if (newRes > res) newRes = res;
       if (newRes > cutEta) cutEta = newRes;
 
       ed = ed->next;
      }

   if ((actTriang->maxDepth) == 0)
     {
      sum = actEstor->globError;
      cutEta = sum*(timeControl->gauge) / (actTriang->noOfEdges);
     }
/*
   else
     if ( (refor->mBar > 1.e-12) && (0.5*(refor->mBar) < cutEta) )  
         cutEta = 0.5*(refor->mBar);
*/
   return cutEta;
}

static real NumAssSigma(int i,int kk)
{
   int  noOfAreaIP = actIntegData->noOfAreaIP;
   real *areaIW = actIntegData->areaIW,
        **areaVals = actIntegData->areaVals;
   real hilfe = ZERO;
   int m;

   for (m = 0; m < noOfAreaIP; m++)
        hilfe += areaIW[m]*sigmaxy[m]*areaVals[i][m]*areaVals[kk][m];
   return hilfe;
}

static real NumAssXi(int i)
{
   int  noOfAreaIP = actIntegData->noOfAreaIP;
   real *areaIW = actIntegData->areaIW,
        **areaVals = actIntegData->areaVals;
   real hilfe = ZERO;
   int m;

   for (m = 0; m < noOfAreaIP; m++)
      {
       hilfe += areaIW[m]*xixy[m]*areaVals[i][m];
      }
   return hilfe;
}

static int AddErr(TD *td)
{
   int i, k;
   real x, y, z, dirVal;
   PT  **pts = &(td->p1);
   EDG **eds = &(td->e1);
 
   actLocalData->t = td;

   if (!ErrAss(td, globA, globB, globMad)) return false;
   for (i=0; i<6; i++)
      {
        if ((eds[i]->boundP)!=DIRICHLET)
          {
            for (k=0; k<4; k++)
                RE(eds[i],BRQ) += globA[i+4][k]*RP(pts[k],BSOL);
            RE(eds[i],BDIAG) += globA[i+4][i+4];
            RE(eds[i],BRHS) += globB[i+4];
			RE(eds[i],ed_mad) += globMad[i+4];
          }
        else
          {
            RE(eds[i],BRQ) = ZERO;
            RE(eds[i],BDIAG) = ONE;
			RE(eds[i],ed_mad) = ONE;
            x = ((eds[i]->p1)->x)+(((eds[i]->p2)->x)-((eds[i]->p1)->x))*HALF;
            y = ((eds[i]->p1)->y)+(((eds[i]->p2)->y)-((eds[i]->p1)->y))*HALF;
            z = ((eds[i]->p1)->z)+(((eds[i]->p2)->z)-((eds[i]->p1)->z))*HALF;
            (actProblem->Dirichlet)(x,y,z,eds[i]->classA,&dirVal,0);
            RE(eds[i],BRHS) = dirVal;  /* sorry, more than once */	  
         }
      }
   return true;
  }

static int ErrAxMulRhsDiag(int isol, int irq, int idiag,int irhs)
{
   int n = 10;
   integData *saveIntegData = actIntegData;
	real tau = (actTimeProblem->tau);

    if (globA==nil)
      {
        globA = ExpMatFull((real**)nil, 0, n, FULL);
        globB = ExpVecFull((real*)nil, 0, n);
        globMad = ExpVecFull((real*)nil, 0, n);
        if ((globA==nil)||(globB==nil)||(globMad==nil)) return false;
      }

   if (errIntegData==nil)
     {
       errIntegData = NewIData(actTimeProblem->integErrType, 10, StdShape, theProblem->symP);
       if (errIntegData==nil) return false;
     }
   actIntegData = errIntegData;

	sigma1 = ONE / (ONE + tau);
	sigma2 = tau * sigma1;

   BSOL = isol; BRQ = irq;
   BRHS = irhs; BDIAG = idiag;
   EdgeSetZero(irq);
   EdgeSetZero(irhs);
   EdgeSetZero(idiag);
   EdgeSetZero(ed_mad);

   if (!ApplyTd(AddErr,all)) return false;

   actIntegData = saveIntegData;
   return true;
}

static int UnMark(TD *t)
{
   ITD(t,iMark) = false;
   return true;
}

static int RefIfEdLessTheta(TD *td)
{		
   EDG *e1=td->e1, *e2=td->e2, *e3=td->e3, *e4=td->e4, *e5=td->e5, *e6=td->e6;

   if ( (RE(e1, ed_res)>=theta) || (RE(e2, ed_res)>=theta) ||
        (RE(e3, ed_res)>=theta) || (RE(e4, ed_res)>=theta) ||
        (RE(e5, ed_res)>=theta) || (RE(e6, ed_res)>=theta) )  
      { 
        if ( ((td->type) != T_GREEN_1)  && ((td->type) != T_GREEN_2A)
          && ((td->type) != T_GREEN_2B) && ((td->type) != T_GREEN_3) )
           {
            ITD(td,iMark) = true;
           }
        else
          {
            ITD(td->father,iMark) = true;
          }
      }

   return true;
}

static int RefMarked(TD *t)
{
   if (ITD(t,iMark))
     { 
       RefTr(t); ITD(t,iMark) = false; 
     }
   return true;
}

static int AdaptEdRef(refMethod *refor)
{
   theta = (refor->sigma)*(refor->mBar);
   ed_res = actEstor->ed_res;

   if (iMark<0)
     {
       iMark =
        GetAccess(nodesState->accTetrahedron,sizeof(int),sizeof(int),"iMark");
       if (iMark<0) return false;
     }
   ApplyTd(UnMark,all);
   ApplyTd(RefIfEdLessTheta, all);
   OpenRef();
   ApplyTd(RefMarked,all);
   return CloseRef(refor->verboseP);
}

static void EdgeSetZero(int index)
{
   EDG *ed = actTriang->firstEdge;

   while (ed!=nil)
     {
       if ((ed->firstSon)==nil) RE(ed,index) = ZERO;
       ed = ed->next;
     }
   return;
}

static int ErrAss(TD *t, real **locA, real *locB, real *locMad)
{
   int i, k, m, rc,
       noOfIPoints = actIntegData->noOfIPoints,
       noOfAreaIP = actIntegData->noOfAreaIP;
   real *x = actLocalData->x, *y = actLocalData->y, *z = actLocalData->z,
        *p11 = actLocalData->p11, *p12 = actLocalData->p12, 
        *p13 = actLocalData->p13, *p21 = actLocalData->p21, 
        *p22 = actLocalData->p22, *p23 = actLocalData->p23,
        *p31 = actLocalData->p31, *p32 = actLocalData->p32, 
        *p33 = actLocalData->p33,
	    *q = actLocalData->q, *g = actLocalData->g,
        **shape = actIntegData->shape,
        **shapeX = actIntegData->shapeX,
        **shapeY = actIntegData->shapeY,
        **shapeZ = actIntegData->shapeZ,
/*
        ***sXX   = actIntegData->sXX,
        ***sXY   = actIntegData->sXY,
        ***sXZ   = actIntegData->sXZ,
        ***sYX   = actIntegData->sYX,
        ***sYY   = actIntegData->sYY,
        ***sYZ   = actIntegData->sYZ,
        ***sZX   = actIntegData->sZX,
        ***sZY   = actIntegData->sZY,
        ***sZZ   = actIntegData->sZZ,
*/
        sXX, sXY, sXZ, sYX, sYY, sYZ, sZX, sZY, sZZ,
       
        *iWeight = actIntegData->integWeight,
        *areaIPX = actIntegData->areaIPX,
        *areaIPY = actIntegData->areaIPY;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33, 
        d, di, s, xm, ym, zm, hilfe,
        *iPointX = actIntegData->integPointX,
        *iPointY = actIntegData->integPointY,
        *iPointZ = actIntegData->integPointZ,
         fVal, fVals[9], cauVals[2], km[512];
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real e1111, e1121, e1131, e2111, e2121, e2131, e3111, e3121, e3131;
   real e1212, e1222, e1232, e2212, e2222, e2232, e3212, e3222, e3232;
   real e1313, e1323, e1333, e2313, e2323, e2333, e3313, e3323, e3333;

   PT *p1=t->p1, *p2=t->p2, *p3=t->p3, *p4=t->p4;
   PT **pts = &(t->p1), *pt1, *pt2, *pt3, *pt4;
   TR   **ts;

   actLocalData->t = t;
   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   e1111 = e11*e11;
   e1121 = e11*e21;
   e1131 = e11*e31;
   e2111 = e1121;
   e2121 = e21*e21;
   e2131 = e21*e31;
   e3111 = e1131;
   e3121 = e21*e31;
   e3131 = e31*e31;
   
   e1212 = e12*e12;
   e1222 = e12*e22;
   e1232 = e12*e32;
   e2212 = e1222;
   e2222 = e22*e22;
   e2232 = e22*e32;
   e3212 = e1232;
   e3222 = e2232;
   e3232 = e32*e32;
   
   e1313 = e13*e13;
   e1323 = e13*e23;
   e1333 = e13*e33;
   e2313 = e1323;
   e2323 = e23*e23;
   e2333 = e23*e33;
   e3313 = e1333;
   e3323 = e2333;
   e3333 = e33*e33;

   d = t->volume*6.0;
   di = 1.0/d;

   for (m=0; m<noOfIPoints; m++)
      {
        x[m] = f11*iPointX[m]+f12*iPointY[m]+f13*iPointZ[m]+(p1->x);
        y[m] = f21*iPointX[m]+f22*iPointY[m]+f23*iPointZ[m]+(p1->y);
        z[m] = f31*iPointX[m]+f32*iPointY[m]+f33*iPointZ[m]+(p1->z);
        rc = (actProblem->Laplace)(x[m],y[m],z[m],t->classA,fVals,0,0);
        if (rc==F_FAILED) return false;
        if (rc==F_IGNORE) return false;

        p11[m] = fVals[0];
        p12[m] = fVals[1];
        p13[m] = fVals[2];
        p21[m] = fVals[3];
        p22[m] = fVals[4];
        p23[m] = fVals[5];
        p31[m] = fVals[6];
        p32[m] = fVals[7];
        p33[m] = fVals[8];

        if ((actProblem->Helmholtz)==nil) q[m] = ZERO;
        else
          {
            rc = (actProblem->Helmholtz)(x[m],y[m],z[m],t->classA,&fVal,0,0);
            if ((rc==F_FAILED)||(rc==F_IGNORE)) q[m] = ZERO;
            else q[m] = fVal;
          }
		km[m] = ZERO;
		rc = (actTimeProblem->K)(x[m],y[m],z[m],t->classA,&km[m],0,0);
        if ((actProblem->Source)==nil) g[m] = ZERO;
        else
          {
            rc =ParabolicSource(x[m],y[m],z[m],t->classA,&fVal,0);
            if (rc==F_FAILED) return false;
            g[m] = (rc==F_IGNORE)?ZERO:fVal;
          }
      }
   for (i=0; i<10; i++)
	  {
		for (k=0; k<10; k++) globA[i][k] = ZERO;
		globB[i] = locMad[i] = ZERO;
	  }

   for (i=4; i<10; i++)
   for (k=0; k<4; k++)
    {
     for (m = 0; m<noOfIPoints; m++)
      {
        sXX = shapeX[i][m]*shapeX[k][m];
        sXY = shapeX[i][m]*shapeY[k][m];
        sXZ = shapeX[i][m]*shapeZ[k][m];
        sYX = shapeY[i][m]*shapeX[k][m];
        sYY = shapeY[i][m]*shapeY[k][m];
        sYZ = shapeY[i][m]*shapeZ[k][m];
        sZX = shapeZ[i][m]*shapeX[k][m];
        sZY = shapeZ[i][m]*shapeY[k][m];
        sZZ = shapeZ[i][m]*shapeZ[k][m];

        hilfe = iWeight[m]*di*
         (
/*
             p11[m]*
                    (e1111*sXX[i][k][m]+e1121*sXY[i][k][m]+e1131*sXZ[i][k][m]   +
                     e2111*sYX[i][k][m]+e2121*sYY[i][k][m]+e2131*sYZ[i][k][m]   +
                     e3111*sZX[i][k][m]+e3121*sZY[i][k][m]+e3131*sZZ[i][k][m])  
          +  p22[m]* 
                    (e1212*sXX[i][k][m]+e1222*sXY[i][k][m]+e1232*sXZ[i][k][m]   +
                     e2212*sYX[i][k][m]+e2222*sYY[i][k][m]+e2232*sYZ[i][k][m]   +
                     e3212*sZX[i][k][m]+e3222*sZY[i][k][m]+e3232*sZZ[i][k][m])  
          +  p33[m]*  
                    (e1313*sXX[i][k][m]+e1323*sXY[i][k][m]+e1333*sXZ[i][k][m]   +
                     e2313*sYX[i][k][m]+e2323*sYY[i][k][m]+e2333*sYZ[i][k][m]   +
                     e3313*sZX[i][k][m]+e3323*sZY[i][k][m]+e3333*sZZ[i][k][m])
 */
             p11[m]*
                    (e1111*sXX+e1121*sXY+e1131*sXZ   +
                     e2111*sYX+e2121*sYY+e2131*sYZ   +
                     e3111*sZX+e3121*sZY+e3131*sZZ)  
          +  p22[m]* 
                    (e1212*sXX+e1222*sXY+e1232*sXZ   +
                     e2212*sYX+e2222*sYY+e2232*sYZ   +
                     e3212*sZX+e3222*sZY+e3232*sZZ)  
          +  p33[m]*  
                    (e1313*sXX+e1323*sXY+e1333*sXZ   +
                     e2313*sYX+e2323*sYY+e2333*sYZ   +
                     e3313*sZX+e3323*sZY+e3333*sZZ)


/*
         +   p12[m]* 
               (e1112*sXX[i][k][m]+e1122*sXY[i][k][m]+e1132*sXZ[i][k][m]   +
                e2112*sYX[i][k][m]+e2122*sYY[i][k][m]+e2132*sYZ[i][k][m]   +
                e3112*sZX[i][k][m]+e3122*sZY[i][k][m]+e3132*sZZ[i][k][m])  
         +   p13[m]* 
               (e1113*sXX[i][k][m]+e1123*sXY[i][k][m]+e1133*sXZ[i][k][m]   +
                e2113*sYX[i][k][m]+e2123*sYY[i][k][m]+e2133*sYZ[i][k][m]   +
                e3113*sZX[i][k][m]+e3123*sZY[i][k][m]+e3133*sZZ[i][k][m])  
         +   p21[m]* 
               (e1211*sXX[i][k][m]+e1221*sXY[i][k][m]+e1231*sXZ[i][k][m]   +
                e2211*sYX[i][k][m]+e2221*sYY[i][k][m]+e2231*sYZ[i][k][m]   +
                e3211*sZX[i][k][m]+e3221*sZY[i][k][m]+e3231*sZZ[i][k][m])  
         +   p23[m]* 
               (e1213*sXX[i][k][m]+e1223*sXY[i][k][m]+e1233*sXZ[i][k][m]   +
                e2213*sYX[i][k][m]+e2223*sYY[i][k][m]+e2233*sYZ[i][k][m]   +
                e3213*sZX[i][k][m]+e3223*sZY[i][k][m]+e3233*sZZ[i][k][m])  
         +   p31[m]*  
               (e1311*sXX[i][k][m]+e1321*sXY[i][k][m]+e1331*sXZ[i][k][m]   +
                e2311*sYX[i][k][m]+e2321*sYY[i][k][m]+e2331*sYZ[i][k][m]   +
                e3311*sZX[i][k][m]+e3321*sZY[i][k][m]+e3331*sZZ[i][k][m])  
         +   p32[m]*  
               (e1312*sXX[i][k][m]+e1322*sXY[i][k][m]+e1332*sXZ[i][k][m]   +
                e2312*sYX[i][k][m]+e2322*sYY[i][k][m]+e2332*sYZ[i][k][m]   +
                e3312*sZX[i][k][m]+e3322*sZY[i][k][m]+e3332*sZZ[i][k][m]) 
*/ 
         );

		hilfe += iWeight[m]*d*q[m]*shape[i][m]*shape[k][m];

        locA[i][k] += sigma2*hilfe +
		              sigma1*iWeight[m]*d*km[m]*shape[i][m]*shape[k][m];
        }
	  locA[i][k] /= scaleData;

/*  printf("i,k,locA[i][k] %d %d    %e\n",i,k,locA[i][k]);  */
    }

   for (i=4; i<10; i++)
     {
      for (m = 0; m<noOfIPoints; m++)
         {
        sXX = shapeX[i][m]*shapeX[i][m];
        sXY = shapeX[i][m]*shapeY[i][m];
        sXZ = shapeX[i][m]*shapeZ[i][m];
        sYX = shapeY[i][m]*shapeX[i][m];
        sYY = shapeY[i][m]*shapeY[i][m];
        sYZ = shapeY[i][m]*shapeZ[i][m];
        sZX = shapeZ[i][m]*shapeX[i][m];
        sZY = shapeZ[i][m]*shapeY[i][m];
        sZZ = shapeZ[i][m]*shapeZ[i][m];

          hilfe = iWeight[m]*di*
           (
/*
              p11[m]*
               (e1111*sXX[i][i][m]+e1121*sXY[i][i][m]+e1131*sXZ[i][i][m]   +
                e2111*sYX[i][i][m]+e2121*sYY[i][i][m]+e2131*sYZ[i][i][m]   +
                e3111*sZX[i][i][m]+e3121*sZY[i][i][m]+e3131*sZZ[i][i][m])  
           +  p22[m]* 
               (e1212*sXX[i][i][m]+e1222*sXY[i][i][m]+e1232*sXZ[i][i][m]   +
                e2212*sYX[i][i][m]+e2222*sYY[i][i][m]+e2232*sYZ[i][i][m]   +
                e3212*sZX[i][i][m]+e3222*sZY[i][i][m]+e3232*sZZ[i][i][m])  
           +  p33[m]*  
               (e1313*sXX[i][i][m]+e1323*sXY[i][i][m]+e1333*sXZ[i][i][m]   +
                e2313*sYX[i][i][m]+e2323*sYY[i][i][m]+e2333*sYZ[i][i][m]   +
                e3313*sZX[i][i][m]+e3323*sZY[i][i][m]+e3333*sZZ[i][i][m]) 
*/
              p11[m]*
               (e1111*sXX+e1121*sXY+e1131*sXZ   +
                e2111*sYX+e2121*sYY+e2131*sYZ   +
                e3111*sZX+e3121*sZY+e3131*sZZ)  
           +  p22[m]* 
               (e1212*sXX+e1222*sXY+e1232*sXZ   +
                e2212*sYX+e2222*sYY+e2232*sYZ   +
                e3212*sZX+e3222*sZY+e3232*sZZ)  
           +  p33[m]*  
               (e1313*sXX+e1323*sXY+e1333*sXZ   +
                e2313*sYX+e2323*sYY+e2333*sYZ   +
                e3313*sZX+e3323*sZY+e3333*sZZ) 

/*
           +  p12[m]* 
               (e1112*sXX[i][i][m]+e1122*sXY[i][i][m]+e1132*sXZ[i][i][m]   +
                e2112*sYX[i][i][m]+e2122*sYY[i][i][m]+e2132*sYZ[i][i][m]   +
                e3112*sZX[i][i][m]+e3122*sZY[i][i][m]+e3132*sZZ[i][i][m])  
           +  p13[m]* 
               (e1113*sXX[i][i][m]+e1123*sXY[i][i][m]+e1133*sXZ[i][i][m]   +
                e2113*sYX[i][i][m]+e2123*sYY[i][i][m]+e2133*sYZ[i][i][m]   +
                e3113*sZX[i][i][m]+e3123*sZY[i][i][m]+e3133*sZZ[i][i][m])  
           +  p21[m]* 
               (e1211*sXX[i][i][m]+e1221*sXY[i][i][m]+e1231*sXZ[i][i][m]   +
                e2211*sYX[i][i][m]+e2221*sYY[i][i][m]+e2231*sYZ[i][i][m]   +
                e3211*sZX[i][i][m]+e3221*sZY[i][i][m]+e3231*sZZ[i][i][m])  
           +  p23[m]* 
               (e1213*sXX[i][i][m]+e1223*sXY[i][i][m]+e1233*sXZ[i][i][m]   +
                e2213*sYX[i][i][m]+e2223*sYY[i][i][m]+e2233*sYZ[i][i][m]   +
                e3213*sZX[i][i][m]+e3223*sZY[i][i][m]+e3233*sZZ[i][i][m])  
           +  p31[m]*  
               (e1311*sXX[i][i][m]+e1321*sXY[i][i][m]+e1331*sXZ[i][i][m]   +
                e2311*sYX[i][i][m]+e2321*sYY[i][i][m]+e2331*sYZ[i][i][m]   +
                e3311*sZX[i][i][m]+e3321*sZY[i][i][m]+e3331*sZZ[i][i][m])  
           +  p32[m]*  
               (e1312*sXX[i][i][m]+e1322*sXY[i][i][m]+e1332*sXZ[i][i][m]   +
                e2312*sYX[i][i][m]+e2322*sYY[i][i][m]+e2332*sYZ[i][i][m]   +
                e3312*sZX[i][i][m]+e3322*sZY[i][i][m]+e3332*sZZ[i][i][m])  
*/
           );
         hilfe += iWeight[m]*d*q[m]*shape[i][m]*shape[i][m];
         locA[i][i] += sigma2*hilfe +
		              sigma1*iWeight[m]*d*km[m]*shape[i][m]*shape[i][m];
		 locB[i] += iWeight[m]*g[m]*shape[i][m];
		 locMad[i] += iWeight[m]*d*km[m]*shape[i][m]*shape[i][m];
       }
      locB[i] *= d/scaleData;
	  locA[i][i] /= scaleData;
	  locMad[i] /= scaleData;
/*   printf("i,locA[i][i], locB[i] %d     %e %e\n",i,locA[i][i],locB[i]); */


     }	

/*      Behandlung der Cauchy-Randbedingungen     */

   ts = &(t->t1);
   for (k=0;k<4;k++)
      { 
        if (ts[k]->boundP == CAUCHY)
          {
            pt1 = pts[(k+1)%4]; 
            pt2 = pts[(k+2)%4];
            pt3 = pts[(k+3)%4];
            pt4 = pts[k];

            s = AreaT(pt1,pt2,pt3);
            
            f11 = (pt2->x)-(pt1->x);
            f12 = (pt3->x)-(pt1->x);
            f21 = (pt2->y)-(pt1->y);
            f22 = (pt3->y)-(pt1->y);
            f31 = (pt2->z)-(pt1->z);
            f32 = (pt3->z)-(pt1->z);

            for (m=0; m<noOfAreaIP; m++)
               {
     		 xm = f11*areaIPX[m]+f12*areaIPY[m]+(pt1->x);
     		 ym = f21*areaIPX[m]+f22*areaIPY[m]+(pt1->y);
     		 zm = f31*areaIPX[m]+f32*areaIPY[m]+(pt1->z);

                 (actProblem->Cauchy)(xm,ym,zm,ts[k]->classA,cauVals,0,0);
                 sigmaxy[m] = sigma2*cauVals[0]/scaleData;
                 xixy[m]    = cauVals[1]/scaleData;
               }


            switch (k)
              {
                          case 0:
                            locA[5][1] += s*NumAssSigma(6,0);
                            locA[5][2] += s*NumAssSigma(6,1);
                            locA[5][3] += s*NumAssSigma(6,2);
                            locA[7][1] += s*NumAssSigma(5,0);
                            locA[7][2] += s*NumAssSigma(5,1);
                            locA[7][3] += s*NumAssSigma(5,2);
                            locA[8][1] += s*NumAssSigma(4,0);
                            locA[8][2] += s*NumAssSigma(4,1);
                            locA[8][3] += s*NumAssSigma(4,2);

                            locA[5][5] += s*NumAssSigma(6,6);
                            locA[7][7] += s*NumAssSigma(5,5);
                            locA[8][8] += s*NumAssSigma(4,4);

                            locB[5] += s*NumAssXi(6);
                            locB[7] += s*NumAssXi(5);
                            locB[8] += s*NumAssXi(4);

                            break;
                          case 1:
                            locA[4][0] += s*NumAssSigma(4,2);
                            locA[4][2] += s*NumAssSigma(4,0);
                            locA[4][3] += s*NumAssSigma(4,1);
                            locA[7][0] += s*NumAssSigma(6,2);
                            locA[7][2] += s*NumAssSigma(6,0);
                            locA[7][3] += s*NumAssSigma(6,1);
                            locA[9][0] += s*NumAssSigma(5,2);
                            locA[9][2] += s*NumAssSigma(5,0);
                            locA[9][3] += s*NumAssSigma(5,1);

                            locA[4][4] += s*NumAssSigma(4,4);
                            locA[7][7] += s*NumAssSigma(6,6);
                            locA[9][9] += s*NumAssSigma(5,5);

                            locB[4] += s*NumAssXi(4);
                            locB[7] += s*NumAssXi(6);
                            locB[9] += s*NumAssXi(5);

                            break;
                          case 2:
                            locA[6][0] += s*NumAssSigma(5,1);
                            locA[6][1] += s*NumAssSigma(5,2);
                            locA[6][3] += s*NumAssSigma(5,0);
                            locA[8][0] += s*NumAssSigma(4,1);
                            locA[8][1] += s*NumAssSigma(4,2);
                            locA[8][3] += s*NumAssSigma(4,0);
                            locA[9][0] += s*NumAssSigma(6,1);
                            locA[9][1] += s*NumAssSigma(6,2);
                            locA[9][3] += s*NumAssSigma(6,0);

                            locA[6][6] += s*NumAssSigma(5,5);
                            locA[8][8] += s*NumAssSigma(4,4);
                            locA[9][9] += s*NumAssSigma(6,6);

                            locB[6] += s*NumAssXi(5);
                            locB[8] += s*NumAssXi(4);
                            locB[9] += s*NumAssXi(6);

                            break;
                          case 3:
                            locA[4][0] += s*NumAssSigma(4,0);
                            locA[4][1] += s*NumAssSigma(4,1);
                            locA[4][2] += s*NumAssSigma(4,2);
                            locA[5][0] += s*NumAssSigma(5,0);
                            locA[5][1] += s*NumAssSigma(5,1);
                            locA[5][2] += s*NumAssSigma(5,2);
                            locA[6][0] += s*NumAssSigma(6,0);
                            locA[6][1] += s*NumAssSigma(6,1);
                            locA[6][2] += s*NumAssSigma(6,2);

                            locA[4][4] += s*NumAssSigma(4,4);
                            locA[5][5] += s*NumAssSigma(5,5);
                            locA[6][6] += s*NumAssSigma(6,6);

                            locB[4] += s*NumAssXi(4);
                            locB[5] += s*NumAssXi(5);
                            locB[6] += s*NumAssXi(6);

                            break;
               }

          }
     }

   return true;
}
