#include <stdio.h>

#include "start.h"
#include "zibutil.h"
#include "fixedsize.h"
#include "commands.h"
#include "tc.h"

extern void InitUser();
extern int RdTri(command*), WrTri(command*);
extern int TimeData(command*), TimeDataFile(command*);
#ifdef DDD
extern int Grape(command*), RdTriHyper(command*);
#else
extern int SReadCmd(command*);
#endif

static char *cmdDefFileName = "kastio.def";

static void SetAddresses();

char *kaskPath = "";
char *globBuf = nil;

static REAL totalTime, lastTime;

int main(int argc, char **argv, char **envp)
  {
	char *initDoFileName = nil;

	globBuf = (char*)ZIBAlloc(256L);
	if (globBuf==nil) { ZIBStdOut("Aua\n"); return -1; }
	totalTime = ZIBSeconds();
	lastTime = totalTime;

	Z_Unix(argc,argv,envp);
	Z_ModSet(d_mainModulID,d_ModVersion,"1.9.3");
	Z_Start(d_mainModulID);
	Z_ModGet(d_mainModulID,"do",&initDoFileName);

	if (!InitCommands(kaskPath, cmdDefFileName,
	                  initDoFileName?initDoFileName:"kastio.startup"))
	  { ZIBStdOut("Jammer\n"); return -1; }
	InitFixedLists(100,(void (*)(int, char*))nil);
	SetLibAddresses(-1);
	InitTriang();
	SetTriAddresses(-1);
	SetGraphAddresses();
	SetProblemAddresses();
	SetSolveAddresses();
	InitProblem("ellkask",100,1,100,SYMMETRIC);
#ifdef DDD
	SetWS(150,48,36,60,120);
#else
	SetWS(150,48,36,120);
#endif
	SetAddresses();
	DefDirIte(); DefEnvChol();
	DefCGMethod();
	DefTimeMatMul();
    DefNonePreCond(); DefBPXPreCond();
    DefEllErrMethod();
	InitTimeProblem(100);
	InitPatientProblem();
	SetStdTProblem();
	CmdMainLoop();

	CloseGraphic(-1);
	totalTime = ZIBSeconds()-totalTime;

	Z_End(d_mainModulID,0);

	return 0;
  }

static int PrTime(command *cmd)
  {
    REAL tt = ZIBSeconds();

	if (ParsCheck(cmd,0,0)) return false;

	sprintf(globBuf, "  Time: %.2f secs, since last call %.2f secs, \n",
			tt-totalTime, tt-lastTime);
	ZIBStdOut(globBuf);
	if (solveState!=nil)
	  {
	    if ((solveState->dirTime)!=ZERO)
		  {
		    sprintf(globBuf,"    acc. direct time:  %.3f secs\n",
					solveState->dirTime);
			ZIBStdOut(globBuf);
		  }
	    if ((solveState->renumbTime)!=ZERO)
		  {
		    sprintf(globBuf,"    acc. renumb. time: %.3f secs\n",
					solveState->renumbTime);
			ZIBStdOut(globBuf);
		  }
	    if ((solveState->assTime)!=ZERO)
		  {
		    sprintf(globBuf,"    acc. assemb. time: %.3f secs\n",
					solveState->assTime);
			ZIBStdOut(globBuf);
		  }
	    if ((solveState->iteTime)!=ZERO)
		  {
		    sprintf(globBuf,"    acc. iterat. time: %.3f secs\n",
					solveState->iteTime);
			ZIBStdOut(globBuf);
		  }
	    if ((solveState->estiTime)!=ZERO)
		  {
		    sprintf(globBuf,"    acc. estimat. time: %.3f secs\n",
					solveState->estiTime);
			ZIBStdOut(globBuf);
		  }
	    if ((solveState->refTime)!=ZERO)
		  {
		    sprintf(globBuf,"    acc. ref. time: %.3f secs\n",
					solveState->refTime);
			ZIBStdOut(globBuf);
		  }
	  }
	sprintf(globBuf,"  Allocated memory: %dbyte\n",allocMem);
	ZIBStdOut(globBuf);

	lastTime = tt;
	return true;
  }

static void SetAddresses()
  {
	SetCommand(10, RdTri);
	SetCommand(11, WrTri);
	SetCommand(18, PrTime);
#ifdef DDD
	SetCommand(20, RdTriHyper);
    SetCommand(54, Grape);
#else
    SetCommand(53, SReadCmd);
#endif
	SetCommand(90, TimeData);
	SetCommand(91, TimeDataFile);

	return;
  }
