#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "commands.h"
#include "appevents.h"
#include "params.h"
#include "tc.h"

static int globIndex = -1;
static real norm;

static int CompareMax(PT*);
static int L2_NormNode(char*);
static int EnergyNormNode(char*);
static int TrueErrorNode(char*); 

void L2_Norm()
  {
	L2_NormForIndex(nodesState->rSol);
	timeControl->norm = norm;

	return;
  }

void L2_NormForIndex(int index)
  {
	globIndex = index;
	norm = ZERO;
	ApplyNode(L2_NormNode, all);
	norm = sqrt(norm);

	return;
  }

static int L2_NormNode(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *massLine = NODE_SPARSE_VEC(v,massIndex);
    char **vs = NODE_NEIGHBOR_VEC(v);
	real localNorm = ZERO;

    for (k=0; k<lng; k++)
      localNorm += massLine[k]*RV(vs[k],globIndex);

	norm += RV(v,globIndex)*localNorm;

	return true;
  }

void Maximum()
  {
	timeControl->norm = ZERO;
	globIndex = nodesState->rSol;
	ApplyP(CompareMax, all);

	return;
  }

static int CompareMax(PT *p)
  {
	real cmp;

	cmp = fabs(RP(p, globIndex));
	if (cmp > timeControl->norm)
	  timeControl->norm = cmp;

	return true;
  }

void EnergyNorm()
  {
	EnergyNormForIndex(nodesState->rSol);
	timeControl->energyNorm = norm;

	return;
  }

void EnergyNormForIndex(int index)
  {
	globIndex = index;
	norm = ZERO;
	ApplyNode(EnergyNormNode, all);
	norm = sqrt(norm);

	return;
  }

static int EnergyNormNode(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *stiffLine = NODE_SPARSE_VEC(v,stiffIndex);
    char **vs = NODE_NEIGHBOR_VEC(v);
	real localNorm = ZERO;

    for (k=0; k<lng; k++)
      localNorm += stiffLine[k]*RV(vs[k],globIndex);

	norm += RV(v,globIndex)*localNorm;

	return true;
  }

void TrueError()
  {
	timeControl->trueError = ZERO;
	globIndex = nodesState->rSol;
	ApplyNode(TrueErrorNode, all);
	timeControl->trueError = sqrt(timeControl->trueError);

	return;
  }

static int TrueErrorNode(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *massLine = NODE_SPARSE_VEC(v,massIndex);
    char **vs = NODE_NEIGHBOR_VEC(v);
	real add, localNorm = ZERO, solVals[4];

    for (k=0; k<lng; k++)
	  {
#ifdef DDD
        (actProblem->Sol) (RV(vs[k],rX),RV(vs[k],rY),RV(vs[k],rZ),CV(vs[k],cClassA),solVals,0);
#else
        (actProblem->Sol) (RV(vs[k],rX),RV(vs[k],rY),CV(vs[k],cClassA),solVals,0);
#endif
		add = massLine[k]*(RV(vs[k],globIndex)-solVals[0]);
		localNorm += add;
	  }
#ifdef DDD
    (actProblem->Sol) (RV(v,rX),RV(v,rY),RV(v,rZ),CV(v,cClassA),solVals,0);
#else
    (actProblem->Sol) (RV(v,rX),RV(v,rY),CV(v,cClassA),solVals,0);
#endif
	timeControl->trueError += fabs((RV(v,globIndex)-solVals[0])*localNorm);

	return true;
  }

static int AddEst(PT *p)
  {
	RP(p, rSol) -= RP(p, rError);
	return true;
  }

void TimeError()
  {
	int saveSol = nodesState->rSol;
	proc saveInitMatMul = (proc)(actMatMul->InitMatMul);
	int saveMinSteps = actItor->minSteps, saveMaxSteps = actItor->maxSteps;

	actItor->eps = 1.0 / (timeControl->erFactor);
	actItor->eps *= actItor->eps;
	AssErrorSide();
    NodeSetZero(rError);
	rSol = nodesState->rSol = rError;
	actMatMul->InitMatMul = (int(*)(int,real**(*)(struct BASIC_ELEM*),
	                                int,real*(*)(struct BASIC_ELEM*)))nil;
	actItor->minSteps = 1;
	actItor->maxSteps = 20;
	
	actItor->verboseP = timeControl->verboseP-1;
    if (actItor->InitMethod) (actItor->InitMethod)(actItor);

	actItor->eps *= actItor->res;
    if (actItor->Method) (actItor->Method)(actItor);
    if (actItor->CloseMethod) (actItor->CloseMethod)(actItor);

	timeControl->noErIte =actItor->iteCount;

	rSol = nodesState->rSol = saveSol;
	actMatMul->InitMatMul = (int(*)(int,real**(*)(struct BASIC_ELEM*),
	                                int,real*(*)(struct BASIC_ELEM*)))saveInitMatMul;

	L2_NormForIndex(rError);
/*	ApplyNode(PrintError,all); */
printf("TimeError: norm=%e, globError=%e\n",norm,sqrt(actEstor->globError));
	timeControl->timeEr = norm;
	if (timeControl->addEsti) ApplyP(AddEst, all);
    SendAppEvent(NewSol, (char*)nil, 0, actItor->name);
	actItor->minSteps = saveMinSteps;
	actItor->maxSteps = saveMaxSteps;

	return;
  }
