#include <stdio.h>
#include <string.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "appevents.h"
#include "commands.h"
#include "params.h"
#include "tc.h"

#define NO_OF_TISSUES  11
#define NO_OF_BLOODDATA 3
#define NO_OF_HYPDATA 3

int rField = -1;

static void CopyField(TRIANGULATION*);
static int IntPolField(EDG*);

#ifdef DDD
static real AbsEfield(COORD_PROTO);
#endif

/* Datenvoreinstellung */
/* ------------------- */

/* Farben der Gewebe in der Graphik */

static int color_preset[] = {1,1,3,3,3,7,7,7,5,4,6}; 
static int dddTransTable[] = {0,2,5,10,8,9};

/* Dichten der Gewebetypen in kg/m^3 */

static real rho_tissue_preset[] = 
{
  0.9e+3,0.9e+3,            /* fat */
  1.0e+3,1.0e+3,1.0e+3,     /* muscle */
  1.6e+3,1.6e+3,1.6e+3,     /* bone   */
  1.0e+3,                   /* intestine */
  1.0e+3,                   /* bladder */
  1.0e+3                    /* tumor */
};

/* Dichte von Blut in kg/m^3 */

static real rho_blood = 1.0e+3;

/* (Dichte * spez. Waerme) der Gewebetypen in W sec/m^3/ Grad C */

static real rc_tissue_preset[] =
{
  2.12e+6,2.12e+6,          /* fat */
  3.72e+6,3.72e+6,3.72e+6,  /* muscle */
  2.25e+6,2.25e+6,2.25e+6,  /* bone   */
  3.81e+6,                  /* intestine */
  3.98e+6,                  /* bladder */
  3.72e+6                   /* tumor */
};

/* (Dichte * spez. Waerme) von Blut in W sec/m^3/ Grad C */

static real rc_blood = 3.72e+6;

/* Thermische Leitfaehigkeit der Gewebetypen in W/m/Grad C */

static real kappa_tissue_preset[] =
{
  0.210,0.210,              /* fat */
  0.642,0.642,0.642,        /* muscle */
  0.436,0.436,0.436,        /* bone   */
  0.550,                    /* intestine */
  0.561,                    /* bladder */
  0.642                     /* tumor */
};

/* Blutfluss der Gewebetypen in m^3/kg/sec */

static real perf_tissue_preset[] =
{
  8.33e-7,8.33e-7,             /* fat */
  3.33e-6,3.33e-6,3.33e-6,     /* muscle */
  8.33e-7,8.33e-7,8.33e-7,     /* bone   */
  5.00e-6,                     /* intestine */
  5.00e-6,                     /* bladder */
  4.17e-6                      /* tumor */                  
/*1.67e-6						*/	     
};

/* Elektrische Leitfaehigkeit der Gewebetypen in S/m */

static real sigma_tissue_preset[] =
{
  0.21,0.21,                    /* fat */
  0.8,0.8,0.8,                  /* muscle */
  0.02,0.02,0.02,               /* bone   */
  0.6,                          /* intestine */
  0.2,                          /* bladder */
  0.8                           /* tumor */
};

/* Feldcharakteristik in V^2/m^2*/

static real sigma_ampl = 0.91007164e+5;

/* Bolustemperatur in Grad C*/

static real temp_bolus = 25.0;

/* Temperatur des arteriellen Blutes in Grad C*/

static real temp_blood = 37.0;

/* Waermetransport an der Haut in W/m^2/Grad C*/

static real transport = 45.0;

static real *rho_tissue = nil, *rc_tissue = nil, *kappa_tissue = nil,
            *perf_tissue = nil, *sigma_tissue = nil;
static real *blood_data = nil, *hyp_data = nil;

static char	*tissueNames[] =
{
  "Fat_top","Fat_bottom","Muscle_left",
  "Muscle_right","Muscle_top","Bone_left",
  "Bone_right","Bone_bottom","Intestine",
  "Bladder","Tumor"
};

static char	*bldDataNames[] = {"rho","rhoXspecHeat","temp"};
static char	*hypDataNames[] = {"fieldFactor","bolusTemp","heatTransSkin"};
static char *tempVarName[] = {"temp"};

static int PatientLaplace(COORD_DEF, int classA, real *fVals,
		                  int equation, int variable)
  {
#ifdef DDD
if (classA>5) printf("Undefined classA = %d\n",classA);
	classA = dddTransTable[classA];
if (classA>10) printf("Error in dddTransTable , result = %d\n",classA);
	fVals[0]	=  kappa_tissue[classA]/1.0e2;		/*	pXX	*/
	fVals[1]	=  0.0;		/*	pXY	*/
	fVals[2]	=  0.0;		/*	pXZ	*/
	fVals[3]	=  0.0;		/*	pYX	*/
	fVals[4]	=  kappa_tissue[classA]/1.0e2;		/*	pYY	*/
	fVals[5]	=  0.0;		/*	pYZ	*/
	fVals[6]	=  0.0;		/*	pZX	*/
	fVals[7]	=  0.0;		/*	pZY	*/
	fVals[8]	=  kappa_tissue[classA]/1.0e2;		/*	pZZ	*/
#else
    fVals[0] = kappa_tissue[classA];
    fVals[1] = ZERO;
    fVals[2] = ZERO;
    fVals[3] = kappa_tissue[classA];
#endif
	return F_CONSTANT;
  }

static int PatientHelmholtz(COORD_DEF, int classA, real *fVals,
		                    int equation, int variable)
  {
	real val;

#ifdef DDD
	classA = dddTransTable[classA];
#endif
	val = rho_tissue[classA] * blood_data[1];
	val *= perf_tissue[classA];
#ifdef DDD
	*fVals = val/1.0e6;
#else
	*fVals = val;
#endif
	return F_CONSTANT;
  }

#ifdef DDD
static real AbsEfield(real x, real y, real z)
{
   TD *tdFirst=actTriang->firstTetrahedron, *tdFather;
   PT *p1, *p2, *p3, *p4;
   real absE2, eData[24];
   real e11r, e11i, e12r, e12i, e13r, e13i;
   real e21r, e21i, e22r, e22i, e23r, e23i;
   real e31r, e31i, e32r, e32i, e33r, e33i;
   real e41r, e41i, e42r, e42i, e43r, e43i;
   real f11, f12, f13, f21, f22, f23, f31, f32, f33;
   real e11, e12, e13, e21, e22, e23, e31, e32, e33;
   real phi1, phi2, phi3, phi4;
   real E1r, E1i, E2r, E2i, E3r, E3i;
   real d;
   int  index, k;

   tdFather = actLocalData->t;
   while (tdFather->depth > 0) tdFather = tdFather->father;
   index = tdFather - tdFirst;
   d = 6.0*(tdFather->volume);
/*
   printf("index %d\n",index);
   for (k=0; k<24; k++)
       printf(" k = %d,    eField[index*24 + k] %e",k,eField[index*24 + k]);
*/
   for (k=0; k<24; k++)
       eData[k] = eField[index*24 + k];


   e11r = eData[0]; e11i = eData[1];
   e12r = eData[2]; e12i = eData[3];
   e13r = eData[4]; e13i = eData[5];

   e21r = eData[6]; e21i = eData[7];
   e22r = eData[8]; e22i = eData[9];
   e23r = eData[10]; e23i = eData[11];

   e31r = eData[12]; e31i = eData[13];
   e32r = eData[14]; e32i = eData[15];
   e33r = eData[16]; e33i = eData[17];

   e41r = eData[18]; e41i = eData[19];
   e42r = eData[20]; e42i = eData[21];
   e43r = eData[22]; e43i = eData[23];

   p1 = tdFather->p1;
   p2 = tdFather->p2;
   p3 = tdFather->p3;
   p4 = tdFather->p4;

   f11 = (p2->x)-(p1->x);
   f12 = (p3->x)-(p1->x);
   f13 = (p4->x)-(p1->x);
   f21 = (p2->y)-(p1->y);
   f22 = (p3->y)-(p1->y);
   f23 = (p4->y)-(p1->y);
   f31 = (p2->z)-(p1->z);
   f32 = (p3->z)-(p1->z);
   f33 = (p4->z)-(p1->z);

   e11 = f22*f33 - f23*f32;
   e12 = f13*f32 - f12*f33;
   e13 = f12*f23 - f13*f22;
   e21 = f23*f31 - f21*f33;
   e22 = f11*f33 - f13*f31;
   e23 = f13*f21 - f11*f23;
   e31 = f21*f32 - f22*f31;
   e32 = f12*f31 - f11*f32;
   e33 = f11*f22 - f12*f21;

   phi2 = ( e11*(x-p1->x) + e12*(y-p1->y) + e13*(z-p1->z) ) / d;
   phi3 = ( e21*(x-p1->x) + e22*(y-p1->y) + e23*(z-p1->z) ) / d;
   phi4 = ( e31*(x-p1->x) + e32*(y-p1->y) + e33*(z-p1->z) ) / d;
   phi1 = ONE - (phi2 + phi3 + phi4);

   E1r = e11r*phi1 + e21r*phi2 + e31r*phi3 + e41r*phi4;
   E2r = e12r*phi1 + e22r*phi2 + e32r*phi3 + e42r*phi4;
   E3r = e13r*phi1 + e23r*phi2 + e33r*phi3 + e43r*phi4;

   E1i = e11i*phi1 + e21i*phi2 + e31i*phi3 + e41i*phi4;
   E2i = e12i*phi1 + e22i*phi2 + e32i*phi3 + e42i*phi4;
   E3i = e13i*phi1 + e23i*phi2 + e33i*phi3 + e43i*phi4;

   absE2  = E1r*E1r + E1i*E1i;
   absE2 += E2r*E2r + E2i*E2i;
   absE2 += E3r*E3r + E3i*E3i;

   return absE2;
}
#endif

static int PatientSource(COORD_DEF, int classA, real *fVal,
					     int equation)
  {
	real val;

#ifdef DDD
	classA = dddTransTable[classA];
#endif

	val = rho_tissue[classA] * blood_data[1];
	val *= perf_tissue[classA];
	val *= blood_data[2];

#ifdef DDD
	val /= 1.0e6;
	val += sigma_tissue[classA] * AbsEfield(COORD_PAR) * hyp_data[0]/1.0e2;
#else
    val += sigma_tissue[classA] * ValueAt(COORD_PAR, rField, actLocalData->t) * hyp_data[0];
#endif
	*fVal = val;
	return F_CONSTANT;
  }

static int PatientCauchy(COORD_DEF, int classA, real *fVals,
		                 int equation, int variable)
  {
	real tau = (actTimeProblem->tau), sigma1, sigma2;

	sigma1 = ONE / (ONE + tau);
	sigma2 = tau * sigma1;
#ifdef DDD
	fVals[0] = hyp_data[2] / 1.0e4;
	fVals[1] = (hyp_data[2]/1.0e4) * hyp_data[1] *sigma2;
#else
	fVals[0] = hyp_data[2];
	fVals[1] = hyp_data[2] * hyp_data[1] *sigma2;
#endif
	return F_CONSTANT;
}

static int PatientDirichlet(COORD_DEF,int classA,real *dirVal,
					        int equation)
  {
	*dirVal = hyp_data[1];
	return F_CONSTANT;
  }

static int PatientUStart(COORD_DEF, int classA, real *fVal,
                         int equation)
  {
	*fVal = blood_data[2];
	return F_CONSTANT;
  }

static int PatientK(COORD_DEF, int classA, real *fVal,
		            int equation, int variable)
  {
#ifdef DDD
	classA = dddTransTable[classA];
	*fVal = rc_tissue[classA]/1.0e6;
#else
	*fVal = rc_tissue[classA];
#endif
	return F_CONSTANT;
  }

int InitPatientProblem()
  {
	int i;

	if (rho_tissue == nil)
	  {
		rho_tissue = (real*)NewParamList((ptr)rho_tissue, "TissueDensity", NO_OF_TISSUES,
		                sizeof(real), tissueNames, T_REAL,0, (char **)nil,
						(ptr)nil, (int (*)(char*,char*,int))nil, (int (*)(char*))nil);
		rc_tissue = (real*)NewParamList((ptr)rc_tissue,"TissueDensXspecHeat",NO_OF_TISSUES,
		                sizeof(real), tissueNames, T_REAL, 0, (char **)nil,
						(ptr)nil ,(int (*)(char*,char*,int))nil, (int (*)(char*))nil);
		kappa_tissue = (real*)NewParamList((ptr)kappa_tissue,"TissueThermConduct",NO_OF_TISSUES,
		                sizeof(real), tissueNames, T_REAL, 0, (char **)nil,
						(ptr)nil, (int (*)(char*,char*,int))nil, (int (*)(char*))nil);
		perf_tissue = (real*)NewParamList((ptr)perf_tissue,"TissuePerfusion",NO_OF_TISSUES,
		                sizeof(real), tissueNames, T_REAL, 0, (char **)nil,
						(ptr)nil, (int (*)(char*,char*,int))nil, (int (*)(char*))nil);
		sigma_tissue = (real*)NewParamList((ptr)sigma_tissue,"TissueElectConduct",NO_OF_TISSUES,
		                sizeof(real), tissueNames, T_REAL, 0, (char **)nil,
						(ptr)nil, (int (*)(char*,char*,int))nil,  (int (*)(char*))nil);
		blood_data = (real*)NewParamList((ptr)blood_data,"BloodData",NO_OF_BLOODDATA,
		                sizeof(real), bldDataNames, T_REAL, 0, (char **)nil,
						(ptr)nil, (int (*)(char*,char*,int))nil, (int (*)(char*))nil);
		hyp_data = (real*)NewParamList((ptr)hyp_data,"HyperthermiaData",NO_OF_HYPDATA,
		                sizeof(real), hypDataNames, T_REAL, 0, (char **)nil,
						(ptr)nil, (int (*)(char*,char*,int))nil, (int (*)(char*))nil);

		for (i = 0; i < NO_OF_TISSUES; i++)
		  {
			if (i<noOfAreaCols) areaCols[i] = color_preset[i];
			rho_tissue[i] = rho_tissue_preset[i];
			rc_tissue[i] = rc_tissue_preset[i] / 1.0e+6;
			kappa_tissue[i] = kappa_tissue_preset[i];
			perf_tissue[i] = perf_tissue_preset[i];
			sigma_tissue[i] = sigma_tissue_preset[i];
		  }

		blood_data[0] = rho_blood;
		blood_data[1] = rc_blood;
		blood_data[2] = temp_blood;

		hyp_data[0] = sigma_ampl;
		hyp_data[1] = temp_bolus;
		hyp_data[2] = transport;
	  }
	if (!SetProblem("patient",tempVarName,
					PatientLaplace,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					PatientHelmholtz,
					PatientSource,
					(int(*)(COORD_PROTO,int,real*,int))nil,
					PatientCauchy,
					PatientDirichlet,
					(int(*)(COORD_PROTO,int,real*,int))nil,
					NumAss,
					NumAssR))
	  return false;
    if (!SetTimeProblem("patient",BANKIP,BANKIP,PatientUStart,PatientK,actProblem)) return false; 
#ifndef DDD
	SetEventProc(TriRead,POS_AFTER, (EventProc)nil,
	             (EventProc)CopyField, "Copie rSol to rField");
#endif	
	return true;
}

static int IntPolField(EDG *ed)
  {
	PT *p1 = ed->p1, *p2 = ed->p2, *pm = ed->pm;
	char *v1, *v2, *vm;

	v1 = &((p1->vec)[(nodesState->startPointNodes)[0]]);
	v2 = &((p2->vec)[(nodesState->startPointNodes)[0]]);
	vm = &((pm->vec)[(nodesState->startPointNodes)[0]]);
    RV(vm,rField) = (RV(v1,rField)+RV(v2,rField))*0.5;

    return true;
  }

static void CopyField(TRIANGULATION *triangul)
  {
    if (rField<0)
	  {
	    rField = GetAccess(nodesState->accNode, sizeof(real), sizeof(real),
		                   "rField (kastio)");
	    if (rField<0) return;
        SetEventProc(RefineEdge,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))IntPolField,
                     "Interpolate rField");
	  }
	NodeAssign(rField, rSol);
	return;
  }
