#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "appevents.h"
#include "commands.h"
#include "params.h"
#include "tc.h"

int Stationary(command *cmd)
  {
	int index = 0, no, k, rc, maxPoints = 50000, verboseP = 1, maxSteps = 20;
	double tau = RMAX, tauSave, iteFactor = 0.25, reqErr;
	char *tp;
	real sigma1, sigma2;

	if (ParsCheck(cmd, 0, 100)) return false;
	if (!TimePreSets()) return false;
	tauSave = actTimeProblem->tau;

	if ((timeControl->compTriang) == nil) return false;
	SelTri(timeControl->compTriang);
	if (!ClearRef(timeControl->compTriang, 0, timeControl->verboseP))
	  return false;	  

	reqErr = (timeControl->ellFactor) * (timeControl->eps);
	reqErr *= timeControl->norm;
	reqErr *= reqErr;

	for (k = 1; k <= (cmd->noOfPars); k++)
	  {
		tp = (cmd->pars)[k];
		index = CheckName(&tp, cmd->names, nameClass);
		if (index==-1)
		  {
			sprintf(globBuf,"TimeControl: '%s' unknown parameter to %s\n",(cmd->pars)[k], (cmd->pars)[0]);
			return false;
		  }
		rc = 0;
		switch (index)
		  {
		case 0:          /* tau */
			k++;
			if (!ReadDoublePar(cmd, k, &tau)) break;
			break;
		case 1:          /* reqErr */
			k++;
			if (!ReadDoublePar(cmd, k, &reqErr)) break;
			break;
		case 2:          /* iteFactor */
			k++;
			if (!ReadDoublePar(cmd, k, &iteFactor)) break;
			break;
		case 3:          /* maxPoints */
			k++;
			if (!ReadIntPar(cmd, k, &maxPoints)) break;
			break;
		case 4:          /* maxSteps */
			k++;
			if (!ReadIntPar(cmd, k, &maxSteps)) break;
			break;
		case 5:          /* verboseP */
			k++;
			if (!ReadIntPar(cmd, k, &verboseP)) break;
			break;
		  }
	  }

	actTimeProblem->tau = tau;
	sigma1 = ONE/(ONE+(actTimeProblem->tau));
	sigma2 = (actTimeProblem->tau)*sigma1;

	solveState->lastOp = op_start;
	solveState->maxSteps = maxSteps;
	solveState->maxPoints = maxPoints;
	solveState->verboseP = verboseP;
	solveState->reqGlobError =reqErr;
	actItor->minSteps = 1;
	actItor->maxSteps = 100;
	actItor->iteFactor = iteFactor;

	sprintf(globBuf,"Fixed problem with sigma1=%4.2f, sigma2=%4.2f, reqErr=%10.3e\n",
	        sigma1,sigma2,sqrt(solveState->reqGlobError));
	ZIBStdOut(globBuf);
	if (!(solveState->Solve)()) return false;

    return true;
  }
