#include <stdio.h>
#include <string.h>
#include <math.h>

#include "zibutil.h"
#include "commands.h"
#include "params.h"
#include "tc.h"

real scaleData = 1.0;
static char	*scaleDataNames[] = {"kappaBar"};

static char *varName[] = {"x"};

static int StdK(COORD_DEF, int classA, real *fVal,
		        int equation, int variable)
  {
	*fVal = ONE;
	return F_CONSTANT;
  }

static int Laplace(COORD_DEF, int classA, real *fVals,
		           int equation, int variable)
  {
#ifdef DDD
	fVals[0]	=  1.0;		/*	pXX	*/
	fVals[1]	=  0.0;		/*	pXY	*/
	fVals[2]	=  0.0;		/*	pXZ	*/
	fVals[3]	=  0.0;		/*	pYX	*/
	fVals[4]	=  1.0;		/*	pYY	*/
	fVals[5]	=  0.0;		/*	pYZ	*/
	fVals[6]	=  0.0;		/*	pZX	*/
	fVals[7]	=  0.0;		/*	pZY	*/
	fVals[8]	=  1.0;		/*	pZZ	*/
#else
    fVals[0] = ONE;
    fVals[1] = ZERO;
    fVals[2] = ZERO;
    fVals[3] = ONE;
#endif
	return F_VARIABLE;
  }

static int InconsSource(COORD_DEF, int classA, real *fVal,
					   int equation)
  {
	*fVal = 30.0;
	return F_CONSTANT;
  }

static int InconsUStart(COORD_DEF, int classA, real *fVal,
					   int equation)
  {
	*fVal = -1.0;
	return F_CONSTANT;
  }

static int InconsDirichlet(COORD_DEF,int classA,real *dirVal,
					      int equation)
  {
	*dirVal = ZERO;
	return F_CONSTANT;
  }

static int JohnsonSource(COORD_DEF, int classA, real *fVal,
					     int equation)
  {
	*fVal = ZERO;
	return F_VARIABLE;
  }

static int JohnsonUStart(COORD_DEF, int classA, real *fVal,
					     int equation)
  {
#ifdef DDD
	*fVal = 3952.847075210474*exp(-(x*x+y*y+z*z)*250.0);
#else
	*fVal = 250.0*exp(-(x*x+y*y)*250.0);
#endif
	return F_VARIABLE;
  }

static int JohnsonDirichlet(COORD_DEF, int classA, real *dirVal,
					  int equation)
  {
	real t = actTimeProblem->time;
	real fact = 1.0/(4.0*t+0.004);

#ifdef DDD
    dirVal[0] = sqrt(fact*fact*fact)*exp(-(x*x+y*y+z*z)*fact);
#else
    dirVal[0] = fact*exp(-(x*x+y*y)*fact);
#endif

	return F_VARIABLE;
  }

static int JohnsonSol(COORD_DEF, int classA, real *fVal,
					  int equation)
  {
	real t = actTimeProblem->time;
	real fact = 1.0/(4.0*t+0.004);

#ifdef DDD
    fVal[0] = sqrt(fact*fact*fact)*exp(-(x*x+y*y+z*z)*fact);
#else
    fVal[0] = fact*exp(-(x*x+y*y)*fact);
#endif
	fVal[1] = ZERO;
	fVal[2] = ZERO;

	return F_VARIABLE;
  }

static real R(real t)
  {
	real f;
	
	f = 0.25*(2.0 + sin((double)(REALPI*t)));
	return f;
  }

static real S(real t)
  {
	real f;
	
	f = 0.25*(2.0 + cos((double)(REALPI*t)));
	return f;
  }

#ifdef DDD
static real T(real t)
  {
	real f;
	
	f = 0.25*(2.0 + cos((double)(REALPI*t)));
	return f;
  }
#endif

#ifdef DDD
static int FlahertySource(COORD_DEF, int classA, real *fVal,
					      int equation)
  {
	real f, vx, vy, vz, vt, g, t = actTimeProblem->time;

	vx = x - R(t);
	vx *= vx;
	vy = y - S(t);
	vy *= vy;
	vz = z - T(t);
	vz *= vz;

	f = 0.8*exp(-80.0*(vx + vy +vz));

	vx = x - R(t); 
	vy = y - S(t);
	vz = z - T(t);
	vt = vx * cos((double)(REALPI*t)) - vy * sin((double)(REALPI*t)) -
	     vz * cos((double)(REALPI*t));
	g  = -25600.0 * ( vx*vx + vy*vy + vz*vz) + 480.0 + REALPI*40.0*vt;

	*fVal = f*g;
	return F_VARIABLE;
  }
#else
static int FlahertySource(COORD_DEF, int classA, real *fVal,
					      int equation)
  {
	real f, vx, vy, vt, g, t = actTimeProblem->time;

	vx = x - R(t);
	vx *= vx;
	vy = y - S(t);
	vy *= vy;

	f = 0.8*exp(-80.0*(vx + vy));

	vx = x - R(t); 
	vy = y - S(t);
	vt = vx * cos((double)(REALPI*t)) - vy * sin((double)(REALPI*t));
	g  = -25600.0 * ( vx*vx + vy*vy) + 320.0 + REALPI*40.0*vt;

	*fVal = f*g;
	return F_VARIABLE;
  }
#endif

#ifdef DDD
static int FlahertyDirichlet(COORD_DEF,int classA,real *dirVal,
					         int equation)
  {
	real t = actTimeProblem->time;
	real vx, vy, vz;

	vx = x - R(t);
	vx *= vx;
	vy = y - S(t);
	vy *= vy;
	vz = z - T(t);
	vz *= vz;

	*dirVal = 0.8*exp(-80.0*(vx + vy + vz));

	return F_VARIABLE;
  }
#else
static int FlahertyDirichlet(COORD_DEF,int classA,real *dirVal,
					         int equation)
  {
	real t = actTimeProblem->time;
	real vx, vy;

	vx = x - R(t);
	vx *= vx;
	vy = y - S(t);
	vy *= vy;

	*dirVal = 0.8*exp(-80.0*(vx + vy));

	return F_VARIABLE;
  }
#endif

#ifdef DDD
static int FlahertyUStart(COORD_DEF, int classA, real *fVal,
					      int equation)
  {
	real vx, vy, vz;

	vx = x - R(ZERO);
	vx *= vx;
	vy = y - S(ZERO);
	vy *= vy;
	vz = z - T(ZERO);
	vz *= vz;

	*fVal = 0.8*exp(-80.0*(vx + vy + vz));
	return F_VARIABLE;
  }
#else
static int FlahertyUStart(COORD_DEF, int classA, real *fVal,
					      int equation)
  {
	real vx, vy;

	vx = x - R(ZERO);
	vx *= vx;
	vy = y - S(ZERO);
	vy *= vy;

	*fVal = 0.8*exp(-80.0*(vx + vy));
	return F_VARIABLE;
  }
#endif

#ifdef DDD
static int FlahertySol(COORD_DEF, int classA, real *fVal,
					   int equation)
  {
	real t = actTimeProblem->time;
	real vx, vy, vz;

	vx = x - R(t);
	vx *= vx;
	vy = y - S(t);
	vy *= vy;
	vz = z - T(t);
	vz *= vz;

    fVal[0] = 0.8*exp(-80.0*(vx + vy + vz));
	fVal[1] = ZERO;
	fVal[2] = ZERO;

	return F_VARIABLE;
  }
#else
static int FlahertySol(COORD_DEF, int classA, real *fVal,
					   int equation)
  {
	real t = actTimeProblem->time;
	real vx, vy;

	vx = x - R(t);
	vx *= vx;
	vy = y - S(t);
	vy *= vy;

    fVal[0] = 0.8*exp(-80.0*(vx + vy));
	fVal[1] = ZERO;
	fVal[2] = ZERO;

	return F_VARIABLE;
  }
#endif

int SetStdTProblem()
  {
	real *adr;
	adr = (real*)NewParamList((ptr)(&scaleData), "scaleData", 1, sizeof(real),
	              scaleDataNames, T_REAL,
				  0, (char **)nil, (ptr)nil,
				  (int (*)(char*,char*,int))nil,
				  (int (*)(char*))nil);
	if (adr==nil) return false;
	scaleData = 1.0;

	if (!SetProblem("incons",varName,
					Laplace,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					InconsSource,
					(int(*)(COORD_PROTO,int,real*,int))nil,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					InconsDirichlet,
					(int(*)(COORD_PROTO,int,real*,int))nil,
					NumAss,
					NumAssR))
	  return false;
    if (!SetTimeProblem("incons",BANKIP,BANKIP,
	                    InconsUStart,
						StdK,
						actProblem)) return false; 

	if (!SetProblem("johnson",varName,
					Laplace,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					JohnsonSource,
					(int(*)(COORD_PROTO,int,real*,int))nil,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					JohnsonDirichlet,
					JohnsonSol,
					NumAss,
					NumAssR))
	  return false;
    if (!SetTimeProblem("johnson",FOLKIP,QUADIP,
	                    JohnsonUStart,
						StdK,
						actProblem)) return false; 

	if (!SetProblem("flaherty",varName,
					Laplace,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					FlahertySource,
					(int(*)(COORD_PROTO,int,real*,int))nil,
					(int(*)(COORD_PROTO,int,real*,int,int))nil,
					FlahertyDirichlet,
					FlahertySol,
					NumAss,
					NumAssR))
	  return false;
    if (!SetTimeProblem("flaherty",BANKIP,BANKIP,
	                    FlahertyUStart,
						StdK,
						actProblem)) return false; 
	return true;
}
