#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "tc.h"

int stiffIndex=-1, massIndex=-1, timeMatMulIndex = -1;

static int staticA, staticB;
static real sigma1, sigma2, staticFac;
static problem *timeProblem = nil, *saveProblem = nil;

static int InitTimeMul(int, real**(*)(BASIC_ELEM*),int,real*(*)(BASIC_ELEM*));
static int AXTimeMul(int,int);
static int CorrSparseLine(char*), CorrTimeSparseLine(char*),
           MulSparseLine(char*), MulStiffLine(char*), AddFac(char*);
static int TimeErrorSource(COORD_PROTO,int,real*,int);
static int TimeErrorDirichlet(COORD_PROTO,int,real*,int);
static void NodeSetInitU(int);
static int SetInitU(char*);
static int RetTimeMatrices(TRIANGULATION*);

void AssErrorSide()
  {
	real tau = (actTimeProblem->tau);

	if ((solveState->curStep)==1)
	  InitTimeMul(-1,actProblem->LocAss,-1,actProblem->LocAssB);

	sigma1 = ONE/(ONE+tau);
	sigma2 = tau*sigma1;

    if (timeProblem==nil)
	  {
	    timeProblem = (problem*)ZIBAlloc(sizeof(problem));
	  }
	timeProblem->Source = TimeErrorSource;
	timeProblem->Dirichlet = TimeErrorDirichlet;
	timeProblem->Cauchy = (int(*)(COORD_PROTO,int,real*,int,int))nil;
	saveProblem = actProblem;
	actProblem = timeProblem;

	NodeSetInitU(nodesState->rRhs);
	NodeSub(nodesState->rRhs,nodesState->rRhs,nodesState->rSol);
	staticA = rError;
	staticB = nodesState->rRhs;
	ApplyNode(MulStiffLine,all);
/* maxPrint = 20; ApplyNode(PrintNode,all); */
	AssRHS(nodesState->rRhs,saveProblem->LocAssB);
	staticFac = 0.5*sigma2;
	ApplyNode(AddFac,all);
	NodeSetBoundZero(nodesState->rRhs);

    if (ApplyNode(CorrTimeSparseLine,all)==false) return;
	
	actProblem = saveProblem;
	return;
  }

static int RetTimeMatrices(TRIANGULATION *triang)
  {
    if (stiffIndex>=0)
	  {
		ReturnSparseMatrix(stiffIndex);
		stiffIndex = -1;
	  }
    if (massIndex>=0)
	  {
		ReturnSparseMatrix(massIndex);
		massIndex = -1;
	  }
    if (timeMatMulIndex>=0)
	  {
		ReturnSparseMatrix(timeMatMulIndex);
		timeMatMulIndex = -1;
	  }
	return true;
  }

static int InitTimeMul(int stiff, real** (*NumAss)(BASIC_ELEM*), int rhs, real* (*NumAssB)(BASIC_ELEM*))
  {
	static int first = true;

	if (first)
	  {
        SetEventProc(TriDeleted,POS_BEFORE,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))RetTimeMatrices,
                     "Return old sparse matrix");
		first = false;
	  }

    if (timeProblem==nil)
	  {
	    timeProblem = (problem*)ZIBAlloc(sizeof(problem));
	  }

	timeProblem->Source = ParabolicSource;
	timeProblem->Dirichlet = actProblem->Dirichlet;
	timeProblem->Cauchy = actProblem->Cauchy;
	timeProblem->Helmholtz = actTimeProblem->K;
	saveProblem = actProblem;
	actProblem = timeProblem;
	AssRHS(nodesState->rRhs,NumAssB);
    actProblem = saveProblem;

	ReturnSparseMatrix(stiffIndex);
	ReturnSparseMatrix(massIndex);
	ReturnSparseMatrix(timeMatMulIndex);
	timeMatMulIndex = MakeSparseMatrix(-1, "Combined matrix");
	stiffIndex = AssembleSparse(-1, "Stiffness matrix",NumAss);
	actProblem = timeProblem;
	massIndex = AssembleSparse(-1, "Mass matrix",NumAss);
	actProblem = saveProblem;
   
	sigma1 = ONE/(ONE+(actTimeProblem->tau));
	sigma2 = (actTimeProblem->tau)*sigma1;

    if (ApplyNode(CorrSparseLine,all)==false) return false;

    if (!(nodesState->compNormP))
      if (!BoundSparse(timeMatMulIndex)) return false;

    return timeMatMulIndex;
  }

static int AXTimeMul(int x, int y)
  {
    staticA = x; staticB = y;
    if (ApplyNode(MulSparseLine,all)==false) return false;
    return true;
  }

static int CorrSparseLine(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *stiffLine = NODE_SPARSE_VEC(v,stiffIndex),
	     *matMulLine = NODE_SPARSE_VEC(v,timeMatMulIndex),
	     *massLine = NODE_SPARSE_VEC(v,massIndex);
 
    RV(v, rRhs) /= scaleData;
	for (k=0; k<lng; k++)
      {
        massLine[k] /= scaleData;
        stiffLine[k] /= scaleData;
		matMulLine[k] = sigma1*massLine[k]+sigma2*stiffLine[k];
      }

    return true;
  }

static int CorrTimeSparseLine(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v), index =IV(v,iIndex);
    real *stiffLine = NODE_SPARSE_VEC(v,stiffIndex),
	     *matMulLine = NODE_SPARSE_VEC(v,timeMatMulIndex),
	     *massLine = NODE_SPARSE_VEC(v,massIndex);
    char **vs = NODE_NEIGHBOR_VEC(v);
 
	if (CV(v,cBoundP)==DIRICHLET)
	  {
	    RV(v,rRhs) = ZERO;
		for (k=0; k<lng; k++)
		  matMulLine[k] = (IV(vs[k],iIndex)==index)?ONE:ZERO;
		return true;
	  }
	for (k=0; k<lng; k++)
      matMulLine[k] = sigma1*massLine[k]+sigma2*stiffLine[k];

    return true;
  }

static int MulSparseLine(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *matMulLine = NODE_SPARSE_VEC(v,timeMatMulIndex),
	     sum=ZERO;
    char **vs = NODE_NEIGHBOR_VEC(v);
    
    for (k=0; k<lng; k++)
      {
        sum += matMulLine[k]*RV(vs[k],staticA);
      }
    RV(v,staticB) = sum;
    return true;
  }

static int MulStiffLine(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    real *stiffLine = NODE_SPARSE_VEC(v,stiffIndex),
	     sum=ZERO;
    char **vs = NODE_NEIGHBOR_VEC(v);

    for (k=0; k<lng; k++)
      {
        sum += stiffLine[k]*RV(vs[k],staticB);
      }
    RV(v,staticA) = sum;
    return true;
  }

int ParabolicSource(COORD_DEF, int classA, real *fVal,
					       int equation)
  {
	real tau = (actTimeProblem->tau), sigma1, sigma2,
	            sourceVal, uVal, kVal, para;

	sigma1 = ONE / (ONE + tau);
	sigma2 = tau * sigma1;
	(saveProblem->Source)(COORD_PAR, classA, &sourceVal, equation);
	(actTimeProblem->K)(COORD_PAR, classA, &kVal, equation, equation);
	para = sigma2*sourceVal;

	switch (timeControl->stepTypeP)
	{
	  case START_STEP:
		(actTimeProblem->UStart) (COORD_PAR, classA, &uVal, equation);
		para += sigma1 * kVal * uVal;
		break;
	  case LATER_STEP:
		para += sigma1*kVal*ValueAtOther(COORD_PAR,
		   (nodesState->startPointNodes)[equation]+nodesState->rSol,
		   actLocalData->t);
		break;
	}
	*fVal = para;
	return F_VARIABLE;
  }

static int TimeErrorSource(COORD_DEF, int classA, real *fVal,
					       int equation)
  {
	real saveTime = actTimeProblem->time,sVal1,sVal2;

	(saveProblem->Source)(COORD_PAR, classA, &sVal1, equation);
	actTimeProblem->time -= actTimeProblem->tau;
	(saveProblem->Source)(COORD_PAR, classA, &sVal2, equation);
	*fVal = sVal1-sVal2;
	actTimeProblem->time = saveTime;

	return F_VARIABLE;
  }

static int TimeErrorDirichlet(COORD_DEF, int classA, real *fVal,
					         int equation)
  {
	*fVal = ZERO;
	return true;
  }

static int SetInitU(char *v)
  {
	real uVal;
#ifdef DDD
	(actTimeProblem->UStart) (RV(v,rX), RV(v,rY), RV(v,rZ), CV(v,cClassA), &uVal, 0);
#else
	(actTimeProblem->UStart) (RV(v,rX), RV(v,rY), CV(v,cClassA), &uVal, CV(v,cEqnNo));
#endif
	RV(v,staticA) = uVal;
	return true;
  }

static int SetU(BASIC_ELEM *t)
  {
	PT **pts = &(t->p1), *p;
	int i, k;

	for (k=0; k<POINTS_SIMPLEX; k++)
	  {
		p = pts[k];
	    for (i=0; i<(nodesState->noOfPointNodes); i++)
	      {
#ifdef DDD
			 RP(p,staticA) = ValueAtOther(p->x, p->y, p->z,
			    (nodesState->startPointNodes)[i]+nodesState->rSol, t);
#else
			 RP(p,staticA) = ValueAtOther(p->x, p->y,
			    (nodesState->startPointNodes)[i]+nodesState->rSol, t);
#endif
		  }
	  }
	return true;
  }

static void NodeSetInitU(int index)
  {
    staticA = index;

	switch (timeControl->stepTypeP)
	{
	  case START_STEP:
		ApplyNode(SetInitU,all);
		break;
	  case LATER_STEP:
#ifdef DDD
		ApplyTd(SetU,all);
#else
		ApplyT(SetU,all);
#endif
	    break;
	}
	return;
  }

static int AddFac(char *v)
  {
	RV(v,staticB) = (CV(v,cBoundP)==DIRICHLET)?ZERO:
	                 staticFac*(RV(v,staticB)/scaleData+RV(v,staticA));
	return true;
  }

void DefTimeMatMul()
  {
	DefMatMulMethod("time",true,InitTimeMul,AXTimeMul,(int(*)())nil);
	return;
  }
