#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "commands.h"
#include "params.h"
#include "appevents.h"
#include "tc.h"

static real maxFactor;

void Scale()
  {
	switch (timeControl->scaling)
	  {
	case ABS:
		break;
	case REL:
		L2_Norm();
		break;
	case MAX:
		Maximum();
		break;
	}
	return;
  }

int ComputeStepsizeReduction()
  {
	real red;

	red = ONE / (timeControl->epsilon);
	red = red * timeControl->safe;
	timeControl->reductionFactor = red;

	return true;
  }

void DoStepsizeReduction()
  {
	timeControl->timeStep = (timeControl->timeStep)*(timeControl->reductionFactor);
	timeControl->noOfReductions++;
	sprintf(globBuf,"TimeControl: %d. stepsize-reduction by factor : %8.2e\n",
	        timeControl->noOfReductions, timeControl->reductionFactor);
	ZIBStdOut(globBuf);
	return;
  }

void NewStepsize()
  {
	real epsilon;

	maxFactor = timeControl->timeStep / timeControl->maxTimeStep;
	epsilon = timeControl->epsilon;
	if (maxFactor > epsilon) epsilon = maxFactor;
	timeControl->tRest = timeControl->tEnd - timeControl->time;
	timeControl->newTimeStep = timeControl->timeStep / epsilon;

	return;
  }

void SelectTimeStep()
  {
	timeControl->timeStep = timeControl->newTimeStep;
	if ((timeControl->timeStep) > (timeControl->tRest)) 
		timeControl->timeStep = timeControl->tRest;

	return;
  }

void ComputeFirstTimeStep()
  {
	real initStep;

	EnergyNorm();

	initStep  = TWO*(timeControl->eps)/(timeControl->energyNorm);
	initStep *= (timeControl->rho)*(timeControl->norm);
	initStep *= initStep;

	timeControl->timeStep = initStep;

	return;
  }
	
int OneStep()
  {
	if ((timeControl->compTriang) == nil) return false;
	SelTri(timeControl->compTriang);
	if (!ClearRef(timeControl->compTriang, 0, timeControl->verboseP))
	  return false;
	timeControl->noOfReductions = 0;

	while ((timeControl->noOfReductions) <= (timeControl->maxRed))
	  {
		if (!ClearRef(timeControl->compTriang,0,timeControl->verboseP))
		  return false;

		solveState->lastOp = op_start;
		solveState->maxSteps = 20;
		solveState->maxPoints = 100000;
		solveState->verboseP = (timeControl->verboseP)-1;
		actItor->minSteps = 1;
		actItor->maxSteps = 100;
		actItor->iteFactor = 0.01;
		if ((timeControl->timeStep) == (timeControl->tRest))
			timeControl->newTime = timeControl->tEnd;
		else
			timeControl->newTime = timeControl->time + timeControl->timeStep;

		actTimeProblem->tau = timeControl->timeStep;
		solveState->reqGlobError = (timeControl->ellFactor) * (timeControl->eps);
		solveState->reqGlobError *= timeControl->norm;
		solveState->reqGlobError *= solveState->reqGlobError;
		if (!ImplicitEulerStep()) return false;
		if ((solveState->breakReason) != 4) return false;
		TimeError();
		if (solveState->iteFail)
		  {
			solveState->breakReason = 4;
			ZIBStdOut("TimeControl: iterative solver failed for time error.\n");
			return false;
		  }
		timeControl->error = (timeControl->timeEr) + sqrt(actEstor->globError);
		if ((actProblem->Sol) != nil) TrueError();

		timeControl->timeEr /= timeControl->norm;
		actEstor->globError /= (timeControl->norm)*(timeControl->norm);
		timeControl->error /= timeControl->norm;
		timeControl->trueError /= timeControl->norm;

		if ((timeControl->timeEr) != ZERO)
		  timeControl->epsilon = exp(log(timeControl->timeEr /
		                                 (timeControl->rho * timeControl->eps)) /
									 (REAL) TWO);
		else
		  timeControl->epsilon = ZERO;
		if ((timeControl->error) < (timeControl->eps))
		  {
    SendAppEvent(NewTimeSol, (char*)nil, 0, "kastio");
			FinishStep();
			return true;
		  }
		ComputeStepsizeReduction();
		DoStepsizeReduction();
	}

	sprintf(globBuf,"TimeControl: More than %d prescripted time-step reductions !\n",(timeControl->maxRed));
	ZIBStdOut(globBuf);
	return false;
  }
			
int ImplicitEulerStep()
  {
	actTimeProblem->time = timeControl->time;
	actTimeProblem->time += actTimeProblem->tau;
/*
	if (timeControl->verboseP)
	  {
		sprintf(globBuf, "TimeControl: reqGlobError %9.3e, t=%e, tau=%e on %s\n",
		        sqrt(solveState->reqGlobError),actTimeProblem->time,
				actTimeProblem->tau,actTriang->name);
		ZIBStdOut(globBuf);
	  }
*/
	timeControl->iteCost = ONE;
	timeControl->estiCost = ONE;

	bpxTau = actTimeProblem->tau;
	if (!(solveState->Solve)())
		return false;
	timeControl->iteCost /= (actTriang->noOfPoints);
	timeControl->estiCost /= (actTriang->noOfPoints);

	return true;
  }

void  FinishStep()
  {
	int k;
	real val;

	timeControl->compTriang = timeControl->resultTriang;
	timeControl->resultTriang = actTriang;
	timeControl->time = timeControl->newTime;
	timeControl->noOfTimeSteps++;
	timeControl->stepTypeP = LATER_STEP;

	k = (timeControl->resultTriang)->noOfPoints;
	timeControl->statPoints += k;
	if (k > (timeControl->maxPoints))
		timeControl->maxPoints = k;
	val = (timeControl->error);
	if (val > (timeControl->maxErrorEst))
		timeControl->maxErrorEst = val;
	val = (timeControl->trueError);
	if (val > (timeControl->maxErrorTrue))
		timeControl->maxErrorTrue = val;
	timeControl->totWork += ((timeControl->iteCost)+
	                         (timeControl->noErIte))*(real)((timeControl->resultTriang)->noOfPoints);

	return;
}

int ProjectOnGrid()
  {
	if ((timeControl->compTriang)==nil) return false;
	actTriang = timeControl->compTriang;

	solveState->lastOp = op_start;
	actTimeProblem->tau = ZERO;
	solveState->reqGlobError = (timeControl->ellFactor) * (timeControl->eps);
	solveState->reqGlobError *= timeControl->norm;
	solveState->reqGlobError *= solveState->reqGlobError;
	if (!ImplicitEulerStep())
		return false;
	timeControl->error = sqrt(actEstor->globError);
	if ((actProblem->Sol)!=nil) TrueError();

	timeControl->timeEr = ZERO;
	timeControl->error /= timeControl->norm;
	timeControl->trueError /= timeControl->norm;
	actEstor->globError /= (timeControl->norm)*(timeControl->norm);

	FinishStep();
	(timeControl->noOfTimeSteps)--;

	return true;
}
