#ifndef __TC__
#define __TC__

#ifdef DDD
#include "triang3.h"
#include "nodes3.h"
#include "assemble3.h"
#include "problem3.h"
#include "solve3.h"
#include "graphic3.h"
#define SPACE_DIM 3
#define BASIC_ELEM TD
#define COORD_DEF real x, real y, real z
#define COORD_PROTO real,real,real
#define COORD_PAR x, y, z
#define POINTS_SIMPLEX 4
#define APPLY_BASIC_ELEM ApplyTd
#else
#include "triang2.h"
#include "nodes.h"
#include "assemble.h"
#include "problem.h"
#include "solve.h"
#include "graphic.h"
#define SPACE_DIM 2
#define BASIC_ELEM TR
#define COORD_DEF real x, real y
#define COORD_PROTO real,real
#define COORD_PAR x, y
#define POINTS_SIMPLEX 3
#define APPLY_BASIC_ELEM ApplyT
#endif

struct TIME_CONTROL {
  int scaling, addEsti, verboseP, showTime;
  int noOfTimeSteps, maxSteps, noOfReductions;
  int stepTypeP, maxRed, noErIte;
  real time, newTime, tEnd, timeStep, newTimeStep, tRest;
  real maxTimeStep, reductionFactor;
  real epsilon, error, timeEr, trueError, norm, energyNorm, eps;
  real safe, rho, ellFactor, erFactor;
  real iteCost, estiCost;
  real maxErrorEst, maxErrorTrue, totWork;
  int statPoints, maxPoints, scalF, inform3D;
  real minScal, maxScal, energLmbda, gauge;
  TRIANGULATION	*compTriang, *resultTriang;
};
typedef struct TIME_CONTROL TIME_CONTROL;

struct TimeProblem {
  char *name;
  int integType, integErrType;
  real tau, time;
  real *solVals;
  int (*UStart)(COORD_PROTO,int,real*,int);
  int (*K)(COORD_PROTO,int,real*,int,int);
  problem *probl;
};
typedef struct TimeProblem TimeProblem;
extern TimeProblem *actTimeProblem;

#define	START_STEP 0
#define LATER_STEP 1

#define	BEGIN 0
#define NOTBEGIN 1

#define OPEN   1
#define CLOSED 0

#define ABS    0
#define REL    1
#define MAX    2

#define HELP   5

extern int TimePreSets();
extern int InitTimeProblem(int);
extern int SetTimeProblem(char*,int,int,int(*)(COORD_PROTO,int,real*,int),
  int(*)(COORD_PROTO,int,real*,int,int),problem*);
extern int SetStdTProblem();
extern void DefTimeMatMul();

extern int ClearRef(TRIANGULATION*,int,int);
extern BASIC_ELEM* FindTriangle(COORD_PROTO, BASIC_ELEM*);
extern real ValueAt(COORD_PROTO, int, BASIC_ELEM*);
extern real ValueAtOther(COORD_PROTO, int, BASIC_ELEM*);
extern void L2_Norm();
extern void L2_NormForIndex(int);
extern void EnergyNorm();
extern void EnergyNormForIndex(int);
extern void Maximum();
extern void TimeError();
extern void TrueError();
extern int ParabolicSource(COORD_PROTO,int,real*,int);
extern int StartValue();
extern real InitialU(PT*, BASIC_ELEM*);

extern void Scale();
extern void NewStepsize();
extern void SelectTimeStep();
extern void ComputeFirstTimeStep();
extern void FinishStep();
extern int OneStep();
extern int ProjectOnGrid();
extern int ImplicitEulerStep();
extern void AssErrorSide();

extern void InitInform();
extern void CloseInform();
extern void InformOnTimeStep();

extern TRIANGULATION *CopyCoarseTriangulation(TRIANGULATION*);
extern BASIC_ELEM *FindBasicElem(real*);

extern int InitPatientProblem();

extern int Stationary(command*);

extern int DefDirIte();

extern TIME_CONTROL* timeControl;
extern int rField, rError;
extern int stiffIndex, massIndex;
extern int aPartner;
extern int NewTimeSol;
extern real scaleData;
#endif
