#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "appevents.h"
#include "commands.h"
#include "params.h"
#include "tc.h"

TIME_CONTROL *timeControl = nil;
TimeProblem *actTimeProblem = nil, *timeProblems=nil;
int rError = -1;
int NewTimeSol = -1;

static FILE	*out1 = nil;

static void	InformOnFail(),InformFiles();
static int TimeProbl(command*),TimeData(command*),
           TimeStepping(command*),InfTimePrb(command*);

static real solVals[4];

static void SetInitTriang()
  {
	timeControl->resultTriang = actTriang;
	timeControl->compTriang = CopyCoarseTriangulation(actTriang);
	if ((timeControl->compTriang)==nil) exit(2);

	return;
  }

int TimePreSets()
  {
	if (actTriang==nil) return false;
	if (actProblem==nil) return false;
	if (timeControl!=nil) return true;

	timeControl = (TIME_CONTROL*)ZIBAlloc((long)sizeof(TIME_CONTROL));
	if (timeControl==nil) return false;

	timeControl->timeStep = 1.0e-3;
	timeControl->newTimeStep = 0.0;
	timeControl->tEnd = 1.0;
	timeControl->tRest = 1.0;
	timeControl->time = 0.0;
	timeControl->newTime = 0.0;
	timeControl->scaling = ABS;
	timeControl->addEsti = true;
	timeControl->verboseP = false;
	timeControl->showTime = 0;
	timeControl->noOfTimeSteps = 0;
	timeControl->maxSteps = 1000;
	timeControl->noOfReductions = 0;
	timeControl->reductionFactor = 1.0;
	timeControl->maxRed = 10;
	timeControl->maxTimeStep = 1000.0;
	timeControl->epsilon = 1.0;
	timeControl->norm = 1.0;
	timeControl->energyNorm = 1.0;
	timeControl->error = 1.0;
	timeControl->timeEr = 1.0;
	timeControl->trueError = 1.0;
	timeControl->eps = 1.0e-1;
	timeControl->safe = 0.7;
	timeControl->rho = 0.18;
	timeControl->ellFactor = 0.72;
	timeControl->erFactor = 5.0;
	timeControl->stepTypeP = START_STEP;
	timeControl->noErIte = 1;
	timeControl->iteCost = ZERO;
	timeControl->estiCost = ZERO;
	timeControl->maxErrorEst = ZERO;
	timeControl->maxErrorTrue = ZERO;
	timeControl->statPoints = 0;
	timeControl->maxPoints = 0;
	timeControl->totWork = ZERO;
	timeControl->minScal = ZERO;
	timeControl->maxScal = ONE;
	timeControl->scalF = 1;
	timeControl->gauge = 2.0;
	timeControl->energLmbda = 1.0;
	timeControl->inform3D = false;
    if (NewTimeSol<0)
      NewTimeSol = DefineEventType("Time solution computed",true);
    if (NewTimeSol<0) return false;

	SetInitTriang();
    if (rError<0)
	  {
	    rError = GetAccess(nodesState->accNode, sizeof(real), sizeof(real),
		                   "rError (kastio)");
	    if (rError<0) return false;
	  }
	return true;
}

static int TimeData(command *cmd)
  {
	int index = 0, no, k, rc;
	double eps;
	char *tp;

	if (ParsCheck(cmd, 0, 2)) return false;
	if (!TimePreSets()) return false;

	if ((cmd->noOfPars)==0)
	  {
		sprintf(globBuf, "TimeControl: state of solution process\n");
		ZIBStdOut(globBuf);
		sprintf(globBuf, "    time-steps  %5d - %5d\n",timeControl->noOfTimeSteps,
		        timeControl->maxSteps);
		ZIBStdOut(globBuf);
		sprintf(globBuf, "    time        %9.3e - %9.3e\n",timeControl->time,
		        timeControl->tEnd);
		ZIBStdOut(globBuf);
		sprintf(globBuf, "    error       %9.3e - %9.3e\n",timeControl->error,
		        timeControl->eps);
		ZIBStdOut(globBuf);
		sprintf(globBuf, "    time-error  %9.3e - %9.3e\n",timeControl->timeEr,
		        (timeControl->rho)*(timeControl->eps));
		ZIBStdOut(globBuf);
		sprintf(globBuf, "    stat-error  %9.3e - %9.3e\n",
		        sqrt(actEstor->globError),
		        (timeControl->ellFactor)*(timeControl->eps));
		ZIBStdOut(globBuf);
		if ((actProblem->Sol)!=nil)
		  {
			sprintf(globBuf, "    true-error  %9.3e - %9.3e\n",
			        timeControl->trueError, timeControl->eps);
			ZIBStdOut(globBuf);
		  }
		sprintf(globBuf, "    tau         %9.3e - %9.3e\n",timeControl->timeStep,
		        timeControl->maxTimeStep);
		ZIBStdOut(globBuf);
		sprintf(globBuf, "    norm        %9.3e\n",timeControl->norm);
		ZIBStdOut(globBuf);
		return true;
	  }
	for (k = 1; k <= (cmd->noOfPars); k++)
	  {
		tp = (cmd->pars)[k];
		index = CheckName(&tp, cmd->names, nameClass);
		if (index==-1)
		  {
			sprintf(globBuf,"TimeControl: '%s' unknown parameter to %s\n",(cmd->pars)[k], (cmd->pars)[0]);
			ZIBStdOut(globBuf);
			return false;
		  }
		rc = 0;
		switch (index)
		  {
		case 0:          /* scale */
			k++;
			if (!ReadIntPar(cmd, k, &no)) break;
			timeControl->scaling = (no>2)?0:no;
			break;
		case 1:          /* maxsteps */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->maxSteps = no;
			break;
		case 2:          /* maxred */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->maxRed = no;
			break;
		case 3:          /* time */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->time = eps;
			break;
		case 4:          /* tend */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->tEnd = eps;
			break;
		case 5:          /* timestep */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->timeStep = eps;
			break;
		case 6:          /* maxtimestep */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->maxTimeStep = eps;
			break;
		case 7:          /* eps */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->eps = eps;
			break;
		case 8:          /* safe */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->safe = eps;
			break;
		case 9:          /* rho */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->rho = eps;
			break;
		case 10:          /* ellfactor */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->ellFactor = eps;
			break;
		case 11:          /* erfactor */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->erFactor = eps;
			break;
		case 12:          /* norm */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->norm = eps;
			break;
		case 13:          /* addesti */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->addEsti = no;
			break;
		case 14:          /* verbose */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->verboseP = no;
			break;
		case 15:          /* showtime */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->showTime = no;
			break;
		case 16:          /* min */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->minScal = eps;
			break;
		case 17:          /* max */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->maxScal = eps;
			break;
		case 18:          /* graphscal */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->scalF = no;
			break;
		case 19:          /* lambda */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->energLmbda = eps;
			break;
		case 20:          /* gauge */
			k++;
			if (ReadDoublePar(cmd, k, &eps)) timeControl->gauge = eps;
			break;
		case 21:          /* inform3d */
			k++;
			if (ReadIntPar(cmd, k, &no)) timeControl->inform3D = no;
			break;
		 }
	  }
	return true;
  }

static void InformFiles()
  {
	if ((timeControl->showTime)>1)
	  {
		if (out1==nil) out1 = fopen("kastio.report","w");
		fprintf(out1,"Problem '%s', geometry from '%s'\n",
		        actTimeProblem->name, actTriang->name);
		fprintf(out1,"   eps=%7f, time=[%e,%e] step=%e\n   ellFactor=%7f erFactor=%7f\n",
		        timeControl->eps,
		        timeControl->time,timeControl->tEnd,timeControl->timeStep,
				timeControl->ellFactor,timeControl->erFactor);

		if ((actProblem->Sol)==nil)
		  {
			fprintf(out1,"        time       tau        error   nodes depth #errIte  exp. CPU-time\n");
			fprintf(out1,"--------------------------------------------------------------------------------\n");
		  }
		else
		  {
			fprintf(out1,"        time       tau        error   true-error nodes depth #errIte  exp. CPU-time\n");
			fprintf(out1,"-------------------------------------------------------------------------------------------\n");
		  }

	  }
	return;
  }
	
static void CloseInformPrint(FILE *out)
  {
	fprintf(out,"Statistics of nodal-points : %5d alltogether; %5d maximum\n",
	        timeControl->statPoints, timeControl->maxPoints);
	fprintf(out,"All over error (estimated) : %11.2e\n",
	        timeControl->maxErrorEst);
	if ((actProblem->Sol)!=nil)
	  {
		fprintf(out,"	       (  true   ) : %11.2e\n",
		        timeControl->maxErrorTrue);
	}

	fprintf(out,"Meanvalue of nodal-points  : %11.0f (w. resp. mem.)\n",
	        ceil(timeControl->statPoints/(real)(timeControl->noOfTimeSteps)));
	fprintf(out,"Mean of iteration work     : %11.2e\n",
	        (timeControl->totWork)/(real)(timeControl->statPoints));

	return;
}

void CloseInform()
  {
	if ((timeControl->showTime)>1)
	  {
		CloseInformPrint(out1);
	  }
	return;
  }

static void InformOnFail()
  {
	ZIBStdOut("TimeControl: elliptic solver failed.\n             Reason: unknown\n");
	return;
  }

static void PrintInform(FILE *out)
  {
	switch ((timeControl->showTime)%2)
	  {
	case 0:
		fprintf(out,"%3d   %8.2e",timeControl->noOfTimeSteps,
		        timeControl->time);
		fprintf(out,"   %8.2e",actTimeProblem->tau);
		if ((actProblem->Sol) == nil)
		  fprintf(out,"   %8.2e",timeControl->error);
		else
		  fprintf(out,"   %8.2e   %8.2e", timeControl->error,
		          timeControl->trueError);
		fprintf(out,"   %4d  %3d",
		        (timeControl->resultTriang)->noOfPoints,
				(timeControl->resultTriang)->maxDepth);
		fprintf(out,"    %2d     %8.2f\n",
		        timeControl->noErIte,solveState->totalTime);
		break;
	case 1:
		fprintf(out," %3d   %8.2e",timeControl->noOfTimeSteps,
		        timeControl->time); 
		fprintf(out,"   %8.2e", actTimeProblem->tau);
		if ((actProblem->Sol) != nil)
		  fprintf(out,"  %8.2e", timeControl->trueError);
		fprintf(out,"  %8.2e  %8.2e",timeControl->error,
		        timeControl->timeEr);
		fprintf(out,"  %8.2e",sqrt(actEstor->globError));
		fprintf(out,"   %4d    %5d     %8.2f\n",
		        (timeControl->resultTriang)->noOfPoints,
				(timeControl->resultTriang)->noOfTriangles,
				solveState->totalTime);
		break;
	}
	return;
}

void InformOnTimeStep()
  {
	PrintInform(stdout);

	if ((timeControl->showTime)>1)
	  {
		fprintf(out1,"%4d  %8.2e", timeControl->noOfTimeSteps, timeControl->time);
		fprintf(out1,"   %8.2e", actTimeProblem->tau);
		if ((actProblem->Sol)==nil)
		  fprintf(out1,"   %8.2e", timeControl->error);
		else
		  fprintf(out1,"   %8.2e   %8.2e", timeControl->error,
		          timeControl->trueError);
		fprintf(out1,"   %4d  %3d",
		        (timeControl->resultTriang)->noOfPoints,
				(timeControl->resultTriang)->maxDepth);
		fprintf(out1,"    %2d     %8.2f\n",
		        timeControl->noErIte, solveState->totalTime);
	  }
	return;
  }

static int TimeStepping(command *cmd)
  {
	if (ParsCheck(cmd, 0, 0)) return false;
	if (!TimePreSets()) return false;
	timeControl->tRest = (timeControl->tEnd) - (timeControl->time);

	InformFiles();
	if ((timeControl->timeStep)==ZERO)
	  {
		if (!ProjectOnGrid())
		  {
			InformOnFail();
			return false;
		  }
		ComputeFirstTimeStep();
		Scale();
		InformOnTimeStep();
	  }
	while (((timeControl->time) < (timeControl->tEnd)) && ((timeControl->noOfTimeSteps) < (timeControl->maxSteps)))
	  {
		if (!OneStep())
          {
			InformOnFail();
			return false;
		  }
		Scale();
		NewStepsize();
		SelectTimeStep();
		InformOnTimeStep();
	}
	if ((timeControl->time) >= (timeControl->tEnd))
		ZIBStdOut("TimeControl: time-end reached.\n");

	if ((timeControl->noOfTimeSteps) >= (timeControl->maxSteps))
		ZIBStdOut("TimeControl: number of time-steps reached.\n");

	CloseInform();

	return true;
  }

int InitTimeProblem(int noOfProblems)
  {
    if (timeProblems==nil)
      timeProblems = (TimeProblem*)ZIBAlloc(noOfProblems*sizeof(TimeProblem));
    if (timeProblems==nil) return nil;
    SetCommand(81, TimeData);
    SetCommand(82, TimeStepping);
    SetCommand(83, TimeProbl);
    SetCommand(84, InfTimePrb);
    SetCommand(86, Stationary);

    return true;
  }

int SetTimeProblem(char *name, int integType, int integErrType,
  int (*UStart)(COORD_PROTO,int,real*,int),
  int (*K)(COORD_PROTO,int,real*,int,int),
  problem *probl)
  {
    int no = AddCmdPar("timeproblem", name);

    if (no>=(theProblem->maxNoOfProblems))
      {
        sprintf(globBuf,"Too many time problems, max is %d\n",
                theProblem->maxNoOfProblems);
        ZIBStdOut(globBuf);
        return false;
      }
    actTimeProblem = &(timeProblems[no]);
    actTimeProblem->name = name;
    actTimeProblem->UStart = UStart;
    actTimeProblem->K = K;
    actTimeProblem->solVals = solVals;
	actTimeProblem->probl = probl;
	actTimeProblem->integType = integType;
	actTimeProblem->integErrType = integErrType;

    return true;
  }

static int TimeProbl(command *cmd)
  {
    int index = 0;
    char *tp;

    if (ParsCheck(cmd,0,1)) return false;

    if ((cmd->noOfPars)==1)
      {
        tp = (cmd->pars)[1];
        index = CheckName(&tp, cmd->names, nameClass);
      }
    if (index==-1)
      {
        sprintf(globBuf,"Parameter %s of %s not allowed\n", (cmd->pars)[1], (cmd->pars)[0]);
        ZIBStdOut(globBuf);
        return false;
      }
    if (timeProblems==nil) return false;
    actTimeProblem = &(timeProblems[index]);
    actProblem = actTimeProblem->probl;
#ifdef DDD
	linIntegData = NewIData(actTimeProblem->integType, 4, StdShape, theProblem->symP);
#else
	linIntegData = NewIData(actTimeProblem->integType, 3, StdShape, theProblem->symP);
#endif
    actIntegData = linIntegData;
	SendAppEvent(ProblemChanged, (char*)actProblem, sizeof(problem), "Problem");
	autoSolEvent = NewGlobSol;

    return true;
  }

static int InfTimePrb(command *cmd)
  {
    int k;

    sprintf(globBuf,"Current problem is '%s' from problem class '%s'\n",
            actProblem->name, theProblem->name);
    ZIBStdOut(globBuf);
    sprintf(globBuf,"Current time problem is '%s'\n",actTimeProblem->name);
    ZIBStdOut(globBuf);
    for (k=0; k<(theProblem->noOfEquations); k++)
      if ((actProblem->varNames)[k]!=nil)
        {
          sprintf(globBuf,"var%d: %s\n",k,(actProblem->varNames)[k]);
          ZIBStdOut(globBuf);
        }
    return true;
  }
