#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "appevents.h"
#include "commands.h"
#include "params.h"
#include "tc.h"

#define MAX_POINT_COLLECT 100

static int AddPointCollect(real x[SPACE_DIM]);
static int CollectValues();
static void FindMinMax(real*,real*,int);
static int MinMaxOnNode(char*);

static real *pointCoord = nil;
static BASIC_ELEM **elem1Collect = nil, **elem2Collect = nil;
static int toCollect = 0;
static FILE *dataFile = nil;
static real globMin, globMax;
static int globIndex;
static TRIANGULATION *refTriang;
static char *globVecMin = nil, *globVecMax = nil;

int TimeData(command *cmd)
  {
	int k;
	real dummy;
	real x[SPACE_DIM];

	if (ParsCheck(cmd, SPACE_DIM, SPACE_DIM)) return false;

	for (k=0; k<SPACE_DIM; k++)
	  if (ReadDoublePar(cmd, k+1, &dummy)) x[k] = dummy;
	  else return false;

	return AddPointCollect(x);
  }

int TimeDataFile(command *cmd)
  {
	if (ParsCheck(cmd, 1, 1)) return false;

	dataFile = fopen((cmd->pars)[1],"w");
	if (dataFile==nil) return false;

	SetEventProc(NewSol,POS_AFTER, (EventProc)nil,
				 (EventProc)CollectValues, "Collect time data");

	return true;
  }

static int AddPointCollect(real x[SPACE_DIM])
  {
	int k, offset;
	BASIC_ELEM *t;

	if (toCollect==MAX_POINT_COLLECT)
	  {
	    ZIBStdOut("Too many points for time data collection\n");
		return false;
	  }
	if (pointCoord==nil)
	  {
		pointCoord = (real*) ZIBAlloc(SPACE_DIM*MAX_POINT_COLLECT*sizeof(real));
		if (pointCoord==nil) return false;
	  }
	if (elem1Collect==nil)
	  {
		refTriang = actTriang;
		elem1Collect = (BASIC_ELEM**) ZIBAlloc(MAX_POINT_COLLECT*sizeof(BASIC_ELEM*));
		elem2Collect = (BASIC_ELEM**) ZIBAlloc(MAX_POINT_COLLECT*sizeof(BASIC_ELEM*));
		if ((elem1Collect==nil)||(elem2Collect==nil)) return false;
	  }
	offset = toCollect*SPACE_DIM;
	for (k=0; k<SPACE_DIM; k++) pointCoord[offset+k] = x[k];
	t = FindBasicElem(x);
	if (t==nil)
	  {
	    ZIBStdOut("AddPointCollect: point not in triangulatio\n");
		return false;
	  }
	if (refTriang==actTriang)
	  {
#ifdef DDD
	    elem1Collect[toCollect] = t;
	    elem2Collect[toCollect] = (BASIC_ELEM*)ATD(t,aPartner);
#else
	    elem1Collect[toCollect] = t;
	    elem2Collect[toCollect] = (BASIC_ELEM*)AT(t,aPartner);
#endif
	  }
	else
	  {
#ifdef DDD
	    elem2Collect[toCollect] = t;
	    elem1Collect[toCollect] = (BASIC_ELEM*)ATD(t,aPartner);
#else
	    elem2Collect[toCollect] = t;
	    elem1Collect[toCollect] = (BASIC_ELEM*)AT(t,aPartner);
#endif
	  }
	toCollect++;

	return true;
  }

static int CollectValues()
  {
	int k;
	real solMin, solMax, x, y, z;
	BASIC_ELEM *t;

	if (dataFile==nil) return true;

	FindMinMax(&solMin, &solMax, rSol);
	fprintf(dataFile,"%7d %13.6e %13.6e %13.6e ",actTriang->noOfPoints,
	        actTimeProblem->time,solMin,solMax);
	
	for (k=0; k< toCollect; k++)
	  {
	    x = pointCoord[k*SPACE_DIM];
	    y = pointCoord[k*SPACE_DIM+1];
	    if (SPACE_DIM>2) z = pointCoord[k*SPACE_DIM+2];
		t = (refTriang==actTriang)?elem1Collect[k]:elem2Collect[k];
		fprintf(dataFile,"%e ", ValueAt(COORD_PAR,rSol,t));
	  }
	fprintf(dataFile,"\n");
	fflush(dataFile);

	return true;
  }

static void FindMinMax(real *minVal, real *maxVal, int ind)
  {
	globMin = RMAX; globMax = RMIN; globIndex = ind;
	ApplyNode(MinMaxOnNode,all);
	*minVal = globMin; *maxVal = globMax;
/*
printf("Minimum %e at (%f,%f,%f), bnd=%d(%d)\n",globMin,
RV(globVecMin,rX),RV(globVecMin,rY),RV(globVecMin,rZ),
CV(globVecMin,cBoundP),CV(globVecMax,cClassA));
printf("Maximum %e at (%f,%f,%f), boundary=%d(%d)\n",globMax,
RV(globVecMax,rX),RV(globVecMax,rY),RV(globVecMin,rZ),
CV(globVecMax,cBoundP),CV(globVecMax,cClassA));
*/
	return;
  }

static int MinMaxOnNode(char *v)
  {
	real value = RV(v, globIndex);

	if (value<globMin) { globMin = value; globVecMin =v; }
	if (value>globMax) { globMax = value; globVecMax =v; }

	return true;
  }

