#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "appevents.h"
#include "commands.h"
#include "tc.h"

#define POSITIV -1.0e-100

extern int RdTri(command*);

int aPartner = -1;

static int SetPartner(BASIC_ELEM*), UnSetPartner(BASIC_ELEM*);
static int UnSetAllPartners(TRIANGULATION*);
static int CheckPartnerCmd(command*);
static int XYinT(BASIC_ELEM*);
#ifdef DDD
static real Jacobi(PT*,PT*,PT*,real*);
#else
static real Jacobi(PT*,PT*,real*);
#endif

static real globX[SPACE_DIM];
BASIC_ELEM *inTriangle, *partnerInTriangle;

TRIANGULATION *CopyCoarseTriangulation(TRIANGULATION *actT)
  {
    TRIANGULATION *saveT = actTriang, *newT;
	BASIC_ELEM *t, *tCopy;
	command cmd;
	char *pars[3];
	int k;

	cmd.no = 10;
	cmd.noOfPars = 1;
	cmd.pars = pars;
	cmd.names=nil;
	cmd.results = nil;
	cmd.verbose = false;
	pars[0] = "readtintern";
	pars[1] = actT->fileName;
	pars[2] = nil;

	if (!(actTriang->Read)(&cmd)) return nil;
	newT = actTriang;
	newT->name = "Kopie";
	SelTri(saveT);

	if (aPartner<0)
	  {
#ifdef DDD
	    aPartner = GetAccess(nodesState->accTetrahedron, sizeof(char*), sizeof(char*),
		                   "aPartner (kastio)");
#else
	    aPartner = GetAccess(nodesState->accTriangle, sizeof(char*), sizeof(char*),
		                   "aPartner (kastio)");
#endif
	    if (aPartner<0) return nil;
		SetCommand(85, CheckPartnerCmd);
	  }

#ifdef DDD
	for (k=0; k<actTriang->noOfInitTetrahedra; k++)
	  {
	    t = &((actTriang->initTetrahedra)[k]);
		tCopy = &((newT->initTetrahedra)[k]);
		ATD(t,aPartner) = (ptr)tCopy;
		ATD(tCopy,aPartner) = (ptr)t;
	  }
	SetEventProc(RefineTetrahedron,POS_AFTER, (EventProc)nil,
	             (EventProc)SetPartner, "Set partner tetrahedron");
	SetEventProc(DeleteTetrahedron,POS_AFTER, (EventProc)nil,
	             (EventProc)UnSetPartner, "Unset partner tetrahedron");
#else
	for (k=0; k<actTriang->noOfInitTriangles; k++)
	  {
	    t = &((actTriang->initTriangles)[k]);
		tCopy = &((newT->initTriangles)[k]);
		AT(t,aPartner) = (ptr)tCopy;
		AT(tCopy,aPartner) = (ptr)t;
	  }
	SetEventProc(RefineTriangle,POS_AFTER, (EventProc)nil,
	             (EventProc)SetPartner, "Set partner triangle");
	SetEventProc(DeleteTriangle,POS_AFTER, (EventProc)nil,
	             (EventProc)UnSetPartner, "Unset partner triangle");
#endif

	return newT;
  }

#ifdef DDD
static int SetPartner(BASIC_ELEM *t)
  {
	BASIC_ELEM *tPartner, *tSons=t->firstSon, *tPartnerSons;
	int depth = t->depth, depthPartner;

	if (aPartner<0) return false;
	tPartner = (BASIC_ELEM*)ATD(t,aPartner);
	if (tPartner==nil)
	  {
	    printf("SetPartner:tPartner nil?\n",tPartner);
printf("t (%d,%d,%d,%d), type =%d\n",
t->p1->indexP,t->p2->indexP,t->p2->indexP,t->p4->indexP,
t->type);
tPartner->p1 = nil;
		return false;
	  }
	depthPartner = tPartner->depth;
	if ((t->type)==T_GREEN_1)
	  {
		ATD(&tSons[0],aPartner) = ATD(&tSons[1],aPartner) = (ptr)tPartner;
		return true;
	  }
	if (((t->type)==T_GREEN_2A)||((t->type)==T_GREEN_3))
	  {
		ATD(&tSons[0],aPartner) = ATD(&tSons[1],aPartner) = (ptr)tPartner;
		ATD(&tSons[2],aPartner) = ATD(&tSons[3],aPartner) = (ptr)tPartner;
		return true;
	  }
	if ((t->type)==T_GREEN_2B)
	  {
		ATD(&tSons[0],aPartner) = ATD(&tSons[1],aPartner) =
		ATD(&tSons[2],aPartner) = (ptr)tPartner;
		return true;
	  }
	if ((depth>depthPartner)||((tPartner->firstSon)==nil)||
	    ((tPartner->type)==T_GREEN_1)||((tPartner->type)==T_GREEN_2A)||
	    ((tPartner->type)==T_GREEN_2B)||((tPartner->type)==T_GREEN_3)) 
	  {
	    ATD(&tSons[0],aPartner) = ATD(&tSons[1],aPartner) =
	    ATD(&tSons[2],aPartner) = ATD(&tSons[3],aPartner) =
	    ATD(&tSons[4],aPartner) = ATD(&tSons[5],aPartner) =
		ATD(&tSons[6],aPartner) = ATD(&tSons[7],aPartner) = (ptr)tPartner;
		return true;
	  }
	tPartnerSons = tPartner->firstSon;
	ATD(&tSons[0],aPartner) = (ptr)(&tPartnerSons[0]);
	ATD(&tSons[1],aPartner) = (ptr)(&tPartnerSons[1]);
	ATD(&tSons[2],aPartner) = (ptr)(&tPartnerSons[2]);
	ATD(&tSons[3],aPartner) = (ptr)(&tPartnerSons[3]);
	ATD(&tSons[4],aPartner) = (ptr)(&tPartnerSons[4]);
	ATD(&tSons[5],aPartner) = (ptr)(&tPartnerSons[5]);
	ATD(&tSons[6],aPartner) = (ptr)(&tPartnerSons[6]);
	ATD(&tSons[7],aPartner) = (ptr)(&tPartnerSons[7]);

	ATD(&tPartnerSons[0],aPartner) = (ptr)(&tSons[0]);
	ATD(&tPartnerSons[1],aPartner) = (ptr)(&tSons[1]);
	ATD(&tPartnerSons[2],aPartner) = (ptr)(&tSons[2]);
	ATD(&tPartnerSons[3],aPartner) = (ptr)(&tSons[3]);
	ATD(&tPartnerSons[4],aPartner) = (ptr)(&tSons[4]);
	ATD(&tPartnerSons[5],aPartner) = (ptr)(&tSons[5]);
	ATD(&tPartnerSons[6],aPartner) = (ptr)(&tSons[6]);
	ATD(&tPartnerSons[7],aPartner) = (ptr)(&tSons[7]);

	return true;
  }
#else
static int SetPartner(TR *t)
  {
        TR *tPartner, *tSons=t->firstSon, *tPartnerSons;
        int depth = t->depth, depthPartner;

        if (aPartner<0) return false;
        tPartner = (TR*)AT(t,aPartner);
        if (tPartner==nil)
          {
            printf("tPartner nil?\n",tPartner);
                return false;
          }
        depthPartner = tPartner->depth;
        if ((t->type)==T_GREEN)
          {
                AT(&tSons[0],aPartner) = AT(&tSons[1],aPartner) = (ptr)tPartner;
                return true;
          }
        if ((depth>depthPartner)||((tPartner->firstSon)==nil)||
            ((tPartner->type)==T_GREEN))
          {
            AT(&tSons[0],aPartner) = AT(&tSons[1],aPartner) =
                AT(&tSons[2],aPartner) = AT(&tSons[3],aPartner) = (ptr)tPartner;
                return true;
          }
        tPartnerSons = tPartner->firstSon;
        AT(&tSons[0],aPartner) = (ptr)(&tPartnerSons[0]);
        AT(&tSons[1],aPartner) = (ptr)(&tPartnerSons[1]);
        AT(&tSons[2],aPartner) = (ptr)(&tPartnerSons[2]);
        AT(&tSons[3],aPartner) = (ptr)(&tPartnerSons[3]);

        AT(&tPartnerSons[0],aPartner) = (ptr)(&tSons[0]);
        AT(&tPartnerSons[1],aPartner) = (ptr)(&tSons[1]);
        AT(&tPartnerSons[2],aPartner) = (ptr)(&tSons[2]);
        AT(&tPartnerSons[3],aPartner) = (ptr)(&tSons[3]);

        return true;
  }
#endif

#ifdef DDD
static int UnSetPartner(BASIC_ELEM *t)
  {
	BASIC_ELEM *tPartner, *tPartnerSons;

	if (((t->type)==T_GREEN_1)||((t->type)==T_GREEN_2A)||
	    ((t->type)==T_GREEN_2B)||((t->type)==T_GREEN_3))
	  return true;

	if (aPartner<0) return false;
	tPartner = (BASIC_ELEM*)ATD(t,aPartner);
	if ((tPartner->type)==T_RED)
	  {
	    tPartnerSons = tPartner->firstSon;
	    ATD(&tPartnerSons[0],aPartner) = ATD(&tPartnerSons[1],aPartner) =
		ATD(&tPartnerSons[2],aPartner) = ATD(&tPartnerSons[3],aPartner) =
		ATD(&tPartnerSons[4],aPartner) = ATD(&tPartnerSons[5],aPartner) =
		ATD(&tPartnerSons[6],aPartner) = ATD(&tPartnerSons[7],aPartner) =
		                                (ptr)t;
	  }
	return true;
  }
#else
static int UnSetPartner(BASIC_ELEM *t)
  {
	BASIC_ELEM *tPartner, *tPartnerSons;

	if ((t->type)==T_GREEN) return true;

	if (aPartner<0) return false;
	tPartner = (BASIC_ELEM*)AT(t,aPartner);
	if ((tPartner->type)==T_RED)
	  {
	    tPartnerSons = tPartner->firstSon;
	    AT(&tPartnerSons[0],aPartner) = AT(&tPartnerSons[1],aPartner) =
		AT(&tPartnerSons[2],aPartner) = AT(&tPartnerSons[3],aPartner) =
		                                (ptr)t;
	  }
	return true;
  }
#endif

#if DDD
static int UnSetAllPartners(TRIANGULATION *triang)
  {
	TRIANGULATION *otherTriang;
	BASIC_ELEM *t;

	otherTriang = (triang==firstTriang)? firstTriang->next: firstTriang;
	t = (otherTriang->depthFirstTetrahedra)[1];
	while (t)
	  {
	    ATD(t,aPartner) = (ptr)nil;
		t = t->next;
	  }
	return true;
  }
#else
static int UnSetAllPartners(TRIANGULATION *triang)
  {
	TRIANGULATION *otherTriang;
	BASIC_ELEM *t;

	otherTriang = (triang==firstTriang)? firstTriang->next: firstTriang;
	t = (otherTriang->depthFirstTriangles)[1];
	while (t)
	  {
	    AT(t,aPartner) = (ptr)nil;
		t = t->next;
	  }
	return true;
  }
#endif

int ClearRef(TRIANGULATION *triang, int depth, int verboseP)
  {
	TRIANGULATION *save = actTriang;

	actTriang = triang;
    if (depth==0)
	  {
		UnSetAllPartners(triang);
		ReturnToCoarse(triang);
		actTriang->maxDepth = 0;
		actTriang->refLevel = 0;
	  }
	else ZIBStdOut("ClearRef: only for depth==0\n");
	actTriang = save;

	return true;
  }

#ifdef DDD
static real Jacobi(PT *p1, PT *p2, PT *p3, real x[SPACE_DIM])
  {
   real x1, x2, x3, y1, y2, y3, z1, z2, z3;

   x1 = (p2->x) - (p1->x);
   x2 = (p2->y) - (p1->y);
   x3 = (p2->z) - (p1->z);

   y1 = (p3->x) - (p1->x);
   y2 = (p3->y) - (p1->y);
   y3 = (p3->z) - (p1->z);

   z1 = x[0] - (p1->x);
   z2 = x[1] - (p1->y);
   z3 = x[2] - (p1->z);

   return x1*(y2*z3 - y3*z2) - x2*(y1*z3 -y3*z1) + x3*(y1*z2 - y2*z1);
  }
#else
static real Jacobi(PT *p1, PT *p2, real x[SPACE_DIM])
  {
    return ((p2->x)-(p1->x))*(x[1]-(p1->y))-
           (x[0]-(p1->x))*((p2->y)-(p1->y));
  }
#endif

#ifdef DDD
real ValueAt(COORD_DEF, int index, BASIC_ELEM *t)
  {
	BASIC_ELEM *tt = t, *ttSons;
	real val, u1, u2, u3, u4,
	     a11, a21, a31, a12, a22, a32, a13, a23, a33,
		 det11, det21, det31, det12, det22, det32, det13, det23, det33,
		 b1, b2, b3,
	     det, alpha, beta, gamma;
	int k, noOfSons = 0;

	globX[0] = x; globX[1] = y; globX[2] = z;
	while ((tt->type)!=T_WHITE)
	{
	  ttSons = tt->firstSon;
	  switch (tt->type)
	    {
	      case T_RED:
		    noOfSons = 8;
			break;
	      case T_GREEN_1:
		    noOfSons = 2;
			break;
	      case T_GREEN_2A:
		    noOfSons = 4;
			break;
	      case T_GREEN_2B:
		    noOfSons = 3;
			break;
	      case T_GREEN_3:
		    noOfSons = 4;
			break;
		  default:
		    noOfSons = 0;
		    break;
		  }

	  for (k=0; k<noOfSons; k++)
	    {
		  if (!XYinT(&ttSons[k]))
		    { tt = &ttSons[k]; break; }
		}
	  if (k==noOfSons)
	    {
		  tt = &ttSons[k-1]; break;
		  ZIBStdOut("ValueAt: no element?\n");
printf("t (%d,%d,%d,%d) type =%d, tt (%d,%d,%d,%d) type =%d\n",
t->p1->indexP,t->p2->indexP,t->p2->indexP,t->p4->indexP,t->type,
tt->p1->indexP,tt->p2->indexP,tt->p2->indexP,
tt->p4->indexP,tt->type);
tt = nil;
		}
	}

	u1 = RP(tt->p1,index);
	u2 = RP(tt->p2,index),
	u3 = RP(tt->p3,index);
	u4 = RP(tt->p4,index);

	a11 = (tt->p2)->x - (tt->p1)->x;
	a21 = (tt->p2)->y - (tt->p1)->y;
	a31 = (tt->p2)->z - (tt->p1)->z;
	a12 = (tt->p3)->x - (tt->p1)->x;
	a22 = (tt->p3)->y - (tt->p1)->y;
	a32 = (tt->p3)->z - (tt->p1)->z;
	a13 = (tt->p4)->x - (tt->p1)->x;
	a23 = (tt->p4)->y - (tt->p1)->y;
	a33 = (tt->p4)->z - (tt->p1)->z;
	
	b1 = x - (tt->p1)->x;
	b2 = y - (tt->p1)->y;
	b3 = z - (tt->p1)->z;
	
	det11 = a22*a33-a32*a23;
	det21 = a12*a33-a32*a13;
	det31 = a12*a23-a22*a13;
	det12 = a21*a33-a31*a23;
	det22 = a11*a33-a31*a13;
	det32 = a11*a23-a21*a13;
	det13 = a21*a32-a31*a22;
	det23 = a11*a32-a31*a12;
	det33 = a11*a22-a21*a12;
	det = a11*det11 - a21*det21 + a31*det31;
	
	alpha  = ( b1*det11 - b2*det21 + b3*det31)/det;
	beta   = (-b1*det12 + b2*det22 - b3*det32)/det;
	gamma  = ( b1*det13 - b2*det23 + b3*det33)/det;
	
	val = u1 + (u2 - u1)*alpha + (u3 - u1)*beta + (u4 - u1)*gamma;
	partnerInTriangle = tt;

	return val;
  }
#else
real ValueAt(COORD_DEF, int index, BASIC_ELEM *t)
  {
	BASIC_ELEM *tt = t;
	EDG *ed;
	real val, u1, u2, u3, a11, a21, a12, a22, b1, b2, det, alpha, beta;

	globX[0] = x; globX[1] = y;
	while ((tt->type)!=T_WHITE)
	  switch (tt->type)
	    {
	      case T_RED:
			if (Jacobi((tt->e3)->pm,(tt->e2)->pm,globX)>=0)
			  { tt = TRED3SON(tt); continue; }
			if (Jacobi((tt->e1)->pm,(tt->e3)->pm,globX)>=0)
			  { tt = TRED1SON(tt); continue; }
			if (Jacobi((tt->e2)->pm,(tt->e1)->pm,globX)>=0)
			  { tt = TRED2SON(tt); continue; }
			tt = TRED4SON(tt);
			continue;
		  case T_GREEN:
			ed = GREENEDGE(tt);
			if (Jacobi(ed->p1,ed->p2,globX)>=0)
			  tt = (tt->firstSon)->next;
			else
			  tt = tt->firstSon;
			break;
		  default:
		    break;
		  }
if (tt->firstSon)
printf("ValueAt: Unexpected son of (%d,%d,%d), type =%d\n",
tt->p1->indexP,tt->p2->indexP,tt->p2->indexP,
tt->type);
	u1 = RP(tt->p1,index);
	u2 = RP(tt->p2,index),
	u3 = RP(tt->p3,index);

	a11 = (tt->p2)->x - (tt->p1)->x;
	a21 = (tt->p2)->y - (tt->p1)->y;
	a12 = (tt->p3)->x - (tt->p1)->x;
	a22 = (tt->p3)->y - (tt->p1)->y;
	
	b1 = x - (tt->p1)->x;
	b2 = y - (tt->p1)->y;
	
	det = a11*a22 - a12*a21;
	
	alpha = (b1*a22 - b2*a12)/det;
	beta  = (a11*b2 - a21*b1)/det;
	
	val = u1 + (u2 - u1)*alpha + (u3 - u1)*beta;
	partnerInTriangle = tt;

	return val;
  }
#endif

#ifdef DDD
static int XYinT(BASIC_ELEM *t)
  {
	if (t->orient)
	  {
		if ((Jacobi(t->p1,t->p2,t->p3,globX)>=POSITIV)&&
			(Jacobi(t->p2,t->p4,t->p3,globX)>=POSITIV)&&
			(Jacobi(t->p3,t->p4,t->p1,globX)>=POSITIV)&&
			(Jacobi(t->p4,t->p2,t->p1,globX)>=POSITIV))
		  { inTriangle = t; return false; }
	  }
	else
	  {
		if ((Jacobi(t->p2,t->p1,t->p3,globX)>=POSITIV)&&
			(Jacobi(t->p1,t->p4,t->p3,globX)>=POSITIV)&&
			(Jacobi(t->p3,t->p4,t->p2,globX)>=POSITIV)&&
			(Jacobi(t->p4,t->p1,t->p2,globX)>=POSITIV))
		  { inTriangle = t; return false; }
	  }
	return true;
  }
#else
static int XYinT(BASIC_ELEM *t)
  {
	if ((Jacobi(t->p1,t->p2,globX)>=0)&&
	    (Jacobi(t->p2,t->p3,globX)>=0)&&
		(Jacobi(t->p3,t->p1,globX)>=0))
	  { inTriangle = t; return false; }
	return true;
  }
#endif

BASIC_ELEM *FindBasicElem(real *x)
  {
	int k;

	inTriangle = nil;
	for (k=0; k<SPACE_DIM; k++) globX[k] = x[k];
	APPLY_BASIC_ELEM(XYinT,all);

	return inTriangle;
  }

real ValueAtOther(COORD_DEF, int index, BASIC_ELEM *t)
  {
#ifdef DDD
    return ValueAt(COORD_PAR, index, (BASIC_ELEM*)ATD(t,aPartner));
#else
    return ValueAt(COORD_PAR, index, (BASIC_ELEM*)AT(t,aPartner));
#endif
  }

static int CheckPartnerCmd(command *cmd)
  {
	double x, y, z = ZERO;
	real val;
	TRIANGULATION *saveTriang = actTriang;
	BASIC_ELEM *t;

	if (ParsCheck(cmd, 2, 3)) return false;
	if (!ReadDoublePar(cmd, 1, &x)) return false;
	if (!ReadDoublePar(cmd, 2, &y)) return false;
	if ((cmd->noOfPars)==3)
	  if (!ReadDoublePar(cmd, 3, &z)) return false;

	actTriang = timeControl->compTriang;
	printf("Looking in triangulation '%s'\n",actTriang->name);
	globX[0]= x; globX[1] = y; globX[2] = z;

	t = FindBasicElem(globX);
	if (t==nil)
	  {
		printf("(%e,%e,%e) not in triangulation\n",
		       globX[0],globX[1],globX[2]);
		actTriang = saveTriang;
		return true;
	  }
#ifdef DDD
	printf("(%e,%e,%e) in triangle (%d,%d,%d,%d)\n",
	       globX[0],globX[1],globX[2],
		   (t->p1)->indexP, (t->p2)->indexP,
		   (t->p3)->indexP,(t->p4)->indexP);
	val = ValueAtOther(globX[0],globX[1],globX[2],rSol,t);
	printf("Partner (%d,%d,%d,%d), val =%e\n",
		   (partnerInTriangle->p1)->indexP,
		   (partnerInTriangle->p2)->indexP,
		   (partnerInTriangle->p3)->indexP,
		   (partnerInTriangle->p4)->indexP,
		   val);
#else
	printf("(%e,%e) in triangle (%d,%d,%d)\n", globX[0], globX[1],
		   (t->p1)->indexP, (t->p2)->indexP,
		   (t->p3)->indexP);
	val = ValueAtOther(globX[0],globX[1],rSol,t);
	printf("Partner (%d,%d,%d), val =%e\n",
		   (partnerInTriangle->p1)->indexP,
		   (partnerInTriangle->p2)->indexP,
		   (partnerInTriangle->p3)->indexP,val);
#endif

	actTriang = saveTriang;
	return true;
  }
